@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 1999-2011  Andrej N. Gritsenko <andrej@rep.kiev.ua>
+ * Copyright (C) 1999-2016  Andrej N. Gritsenko <andrej@rep.kiev.ua>
  *
  *     This program is free software; you can redistribute it and/or modify
  *     it under the terms of the GNU General Public License as published by
@@ -45,6 +45,7 @@ struct peer_t
   struct connchain_i *connchain;	/* connchain instance */
   const char *network_type;		/* for connchain identification */
   struct peer_priv *priv;		/* session-specific data, NULL in thread */
+  struct NODE *modules_data;		/* see PeerData_Attach(), etc. */
   time_t last_input;
   _peer_state state;
   userflag uf;				/* global+direct flags */
@@ -70,6 +71,7 @@ int Connect_Host (const char *, unsigned short, pthread_t *, idx_t *,
 
 int Connchain_Grow (struct peer_t *, char);
 int Connchain_Check (struct peer_t *, char);
+int Connchain_Shrink (struct peer_t *, struct connchain_i *);
 ssize_t Connchain_Put (struct connchain_i **, idx_t, const char *, size_t *)
 			__attribute__((warn_unused_result));
 ssize_t Connchain_Get (struct connchain_i **, idx_t, char *, size_t)
@@ -79,6 +81,11 @@ ssize_t Connchain_Get (struct connchain_i **, idx_t, char *, size_t)
 #define Peer_Put(peer,buf,s) Connchain_Put(&peer->connchain,peer->socket,buf,s)
 #define Peer_Get(peer,buf,s) Connchain_Get(&peer->connchain,peer->socket,buf,s)
 
+int PeerData_Attach (struct peer_t *, const char *, void *, void (*)(void *));
+void PeerData_Detach (struct peer_t *, const char *);
+void *PeerData_Get (struct peer_t *, const char *);
+void Peer_Cleanup (struct peer_t *peer);
+
 /* interop direct.c -> connchain.c */
 void _fe_init_connchains (void);
 
