@@ -177,7 +177,7 @@ void free ();
  */
 
 #define LNAMELEN	32	/* Lname max width */
-#define NAMEMAX		63	/* single name max width, must be >= LNAMELEN */
+#define NAMEMAX		MB_LEN_MAX*LNAMELEN /* single name max width */
 #define PASSWDMAX	NAMEMAX	/* still reserved 3 bytes for connchain flags */
 #define IFNAMEMAX	2*NAMEMAX+1 /* interface name length */
 #define IDENTLEN	10	/* user login name max width */
@@ -481,38 +481,24 @@ typedef short idx_t;
 /* module "ircd" defines */
 
 #define A_SERVER	A_ME	/* no other usage on it */
-#define A_SERVICE	A_ADMIN	/* for umode only */
+#define A_SERVICE	A_SECRET /* for umode only */
+#define A_SSL		A_PRIVATE /* special SSL support, for umode only */
 
 #define IRCDLMAX	32	/* maximum of ircd listeners */
 #define IRCD_DEFAULT_SENDQ 2000	/* default if client is out of any class */
 #define AWAYLEN		120	/* away message max width */
 #define SERVEROPTIONLEN	132	/* see RFC2813 for PASS command */
 #define SERVERNAMELEN	63	/* see RFC2812 */
-#define TOPICLEN	120	/* channel topic max width */
+#define TOPICLEN	255	/* channel topic max width */
 #define KEYLEN		23	/* channel key max width */
 #define MAXMODES	3	/* max changes per MODE command */
-#define MAXCHANNELS	20	/* max channels user can join */
-#define MAXMATCHES	500	/* max number of replies on LIST or NAMES */
-#define MAXTARGETS	4	/* max num. of targets for PRIVMSG or NOTICE */
-#define MAXWHOIS	3	/* max targets to show on WHOIS */
 #define MOTDMAXSIZE	5000	/* size of motd file we think it's not trash */
 
 #undef USE_SERVICES		/* define it if you want to use local services */
 
-  /* send SQUIT to youngest link on duplicate SERVER rather than to announcer */
-#define IRCD_SQUIT_YOUNGEST 1
-#define IRCD_P_FLAG		/* should require 'P' flag from links if def. */
-#undef IRCD_STATM_EMPTYTOO	/* define to show commands never used (stat m) */
-#define IRCD_TRACE_USERS 1	/* define if want opers see all users in TRACE */
-  /* define to grant access to topic outside of channel which have neither mode
-     +n nor +t. note: topic is still partially accessible with LIST command */
-#define IRCD_PUBLIC_TOPIC 1
-#undef IDLE_FROM_MSG		/* define if only PRIVMSG can reset idle time */
-#define DEFAULT_INVISIBLE	/* define to set user mode +i on registration */
-#undef WALLOP_ONLY_OPERS	/* define if only OPERs can get WALLOP messages */
-#undef NO_SPARE_INVITES		/* define to disallow set +I for -i channels */
-#define IRCD_STRICT_MODECMD 1	/* define to disable use such as 'MODE #a n' */
-#undef IRCD_IGNORE_MKEY_ARG	/* define to not compare arg of -k with key */
+#undef IRCD_P_FLAG		/* should require 'P' flag from links if def. */
+
+#define TOPICWHOTIME 1		/* define to send RPL_TOPICWHOTIME to users */
 
   /* define to enable using these commands below by operators */
 #define IRCD_ENABLE_KILL 1
@@ -532,7 +518,7 @@ typedef short idx_t;
 #define IRCD_ID_HISTORY 8192	/* id bitmap size, must be power of 2 */
 
   /* Settings below are Network-Wide, never change them on single server,
-     as well as CHARSET_8BIT, NICKLEN, and CHANNAMELEN! */
+     as well as KEYLEN, MAXMODES, CHARSET_8BIT, NICKLEN, and CHANNAMELEN! */
 #define CHIDLEN		5	/* id len on ! channels */
 #define CHASETIMELIMIT	90	/* "nick delay" for RFC2813 servers */
 #define REOP_DELAY	5400	/* opless timelimit for +r channels */
@@ -694,18 +680,20 @@ typedef short idx_t;
 #define BINDING_TYPE_irc_quit(func) static\
   int func(INTERFACE *, char *, size_t )
 #define BINDING_TYPE_ircd_auth(func) static\
-  int func(struct peer_t *, char *, char *, const char **)
+  int func(struct peer_t *, char *, char *, const char **,\
+	   modeflag *)
 #define BINDING_TYPE_ircd_register_cmd(func) static\
   int func(INTERFACE *, struct peer_t *, int , const char **)
 #define BINDING_TYPE_ircd_client_cmd(func) static\
-  int func(INTERFACE *, struct peer_t *, char *, char *,\
-	   char *, int , const char **)
+  int func(INTERFACE *, struct peer_t *, const char *,\
+	   const char *, const char *, const char *,\
+	   modeflag , int , const char **)
 #define BINDING_TYPE_ircd_server_cmd(func) static\
   int func(INTERFACE *, struct peer_t *, unsigned short ,\
-	   const char *, const char *, char *, int ,\
+	   const char *, const char *, int ,\
 	   const char **)
 #define BINDING_TYPE_ircd_local_client(func) static\
-  void func (INTERFACE *, struct peer_t *)
+  void func (INTERFACE *, struct peer_t *, modeflag )
 #define BINDING_TYPE_ircd_client(func) static\
   void func (INTERFACE *, const char *, const char *,\
 		   const char *, const char *, const char *,\
@@ -721,18 +709,21 @@ typedef short idx_t;
 #define BINDING_TYPE_ircd_check_message(func) static\
   int func(modeflag , modeflag , char *)
 #define BINDING_TYPE_ircd_check_modechange(func) static\
-  int func(modeflag , modeflag , int , modeflag ,\
-	   char *)
+  int func(INTERFACE *, modeflag , modeflag , int ,\
+	   modeflag , char *, modeflag , modeflag )
 #define BINDING_TYPE_ircd_modechange(func) static\
-  modeflag func(INTERFACE *, const char *, modeflag ,\
-		      const char *, modeflag , int ,\
-		      char ,\
+  modeflag func(modeflag , modeflag ,\
+		      const char *, modeflag ,\
+		      modeflag , int , char ,\
 		      int (**)(INTERFACE *, const char *,\
 				 const char *, int ,\
 				 const char **))
 #define BINDING_TYPE_ircd_umodechange(func) static\
-  modeflag func(INTERFACE *, const char *, modeflag ,\
-		      int )
+  modeflag func(modeflag , int ,\
+		      void (**)(INTERFACE *, const char *,\
+				  char *, const char *,\
+				  size_t , int ,\
+				  const char *))
 #define BINDING_TYPE_ircd_whochar(func) static\
   char func(char )
 #define BINDING_TYPE_ircd_client_filter(func) static\
@@ -741,12 +732,19 @@ typedef short idx_t;
 #define BINDING_TYPE_ircd_stats_reply(func) static\
   void func(INTERFACE *, const char *, modeflag )
 #define BINDING_TYPE_ircd_collision(func) static\
-  char *func(char *, size_t , int )
+  char *func(INTERFACE *, char *, size_t , int ,\
+		   const char *, const char *)
 #define BINDING_TYPE_ircd_set_message_targets(func) static\
-  int func (INTERFACE *srv, const char *sender, const char *target,\
-	    int (*next)(INTERFACE *srv, const char *mask,\
-			const char **lcname, modeflag *mf, int **mark,\
-			void **iter))
+  int func (INTERFACE *, const char *, const char *,\
+	    int (*)(INTERFACE *, const char *,\
+			const char **, modeflag *, int **,\
+			void **))
+#define BINDING_TYPE_ircd_whois(func) static\
+  void func(INTERFACE *, const char *, modeflag ,\
+		  const char *, const char *, const char *,\
+		  modeflag )
 #define BINDING_TYPE_ircd_check_send(func) static\
-  int func(INTERFACE *srv, struct peer_t *tgt, modeflag tum, char *msg,\
-	   size_t ms)
+  int func(INTERFACE *, struct peer_t *, modeflag , char *,\
+	   size_t )
+#define BINDING_TYPE_ircd_isupport(func) static\
+  void func(char *, size_t )
