@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 1999-2010  Andrej N. Gritsenko <andrej@rep.kiev.ua>
+ * Copyright (C) 1999-2016  Andrej N. Gritsenko <andrej@rep.kiev.ua>
  *
  *     This program is free software; you can redistribute it and/or modify
  *     it under the terms of the GNU General Public License as published by
@@ -20,7 +20,7 @@
 
 #ifndef INIT_C
 
-WHERE const char *_VERSION	INITVAL("0.10.2");
+WHERE const char *_VERSION	INITVAL("0.11.0");
 
 	/* public part */
 typedef char bool;
@@ -57,6 +57,15 @@ int RegisterBoolean (const char *, bool *);
 int RegisterString (const char *, char *, size_t, int);
 int UnregisterVariable (const char *);
 
+typedef enum {
+  VARIABLE_INTEGER = 0,
+  VARIABLE_BOOLEAN,
+  VARIABLE_CONSTANT,
+  VARIABLE_STRING
+} VariableType;
+
+int GetVariable (const char *, VariableType, void **);
+
 char *SetFormat (const char *, char *);		/* name, new value */
 int Save_Formats (void);
 short *FloodType (const char *);
@@ -107,7 +116,7 @@ void Status_Connchains (INTERFACE *);		/* the same (connchain.c) */
 #endif /* ifndef INIT_C */
 #ifndef DISPATCHER_C
 String  ("nick", Nick, "")			/* default nick and botnetnick */
-Const   ("version", "0.10.2")
+Const   ("version", "0.11.0")
 String  ("my-hostname", hostname, "")
 String  ("locale", locale, "")
 String	("charset", Charset, "")		/* used by couple of modules */
