@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 1999-2014  Andrej N. Gritsenko <andrej@rep.kiev.ua>
+ * Copyright (C) 1999-2016  Andrej N. Gritsenko <andrej@rep.kiev.ua>
  *
  *     This program is free software; you can redistribute it and/or modify
  *     it under the terms of the GNU General Public License as published by
@@ -188,8 +188,6 @@ static int _addhost (user_hr **hr, const char *uh)
   user_hr *h = *hr;
 
   sz = safe_strlen (uh) + 1;
-  if (sz < 6)		/* at least *!i@* :) */
-    return 0;
   *hr = safe_malloc (sz + sizeof(user_hr) - sizeof(h->hostmask));
   _R_h += sz + sizeof(user_hr) - sizeof(h->hostmask);
   memcpy ((*hr)->hostmask, uh, sz);
@@ -251,6 +249,72 @@ static struct clrec_t *_findthebest (const char *mask, struct clrec_t *prefer)
   return user;
 }
 
+/*--- R --- UFLock read --- no HLock ---*/
+static struct clrec_t *_findbyhost (const char *mask)
+{
+  struct clrec_t *u;
+  user_hr *hr = NULL;
+  lid_t lid;
+  char lcmask[HOSTMASKLEN+1];
+
+  unistrlower (lcmask, mask, sizeof(lcmask));
+  rw_rdlock (&HLock);
+  lid = LID_MIN;
+  do {
+    if ((u = UList[lid - LID_MIN]) && !(u->flag & (U_SPECIAL|U_ALIAS)))
+    {
+      /* pseudo-users hosts are not masks */
+      for (hr = u->host; hr; hr = hr->next)
+	if (strcmp (hr->hostmask, lcmask) == 0)
+	  break;
+      if (hr)
+	break;
+    }
+    u = NULL;
+  } while (lid++ != LID_MAX);
+  rw_unlock (&HLock);
+  return u;
+}
+
+static int _check_subpattern(char **ptr, char end1, char end2)
+{
+  char *c;
+  int r = 0, r1, r2;
+
+  for (c = *ptr; *c && *c != end1 && *c != end2; c++)
+  {
+    if (*c == '*' || *c == '?') /* wildcard */
+      continue;
+    if (*c == '[') /* range */
+    {
+      if (*c == '-')
+	c++;
+      if (*c == ']')
+	c++;
+      while (*c && *c != ']')
+	c++;
+      r++;
+    }
+    else if (*c == '{') /* list */
+    {
+      r1 = 1024;
+      for (c++; *c; c++)
+      {
+	r2 = _check_subpattern(&c, ',', '}');
+	if (r1 > r2)
+	  r1 = r2;
+	if (*c == '}')
+	  break;
+      }
+      r += r1; /* add minimum sequence size */
+    }
+    else /* non-wildcard */
+      r++;
+  }
+  *ptr = c;
+  return r;
+}
+
 /*--- W --- no HLock ---*/
 static int _add_usermask (struct clrec_t *user, const char *mask)
 {
@@ -258,12 +322,21 @@ static int _add_usermask (struct clrec_t *user, const char *mask)
   user_hr **h;
   int r;
   char lcmask[HOSTMASKLEN+1];
+  char *c = lcmask;
 
   if (user->uid > ID_ANY && strchr(user->lname, '.'))
     /* it's server name rather than user name */
     strfcpy (lcmask, mask, sizeof(lcmask));
   else
     unistrlower (lcmask, mask, sizeof(lcmask));
+  /* test the pattern to contain not just wildcards */
+  r = _check_subpattern(&c, '!', '@');
+  if (*c == '!')
+    c++, r += _check_subpattern(&c, '@', 0);
+  if (*c == '@')
+    c++, r += _check_subpattern(&c, 0, 0);
+  if (r < 1)			/* at least a single non-wildcard char required */
+    return 0;
   /* check for aliases */
   if (user->flag & U_ALIAS)	/* no need lock since threads has R/O access */
     user = user->u.owner;
@@ -661,6 +734,17 @@ int Add_Clientrecord (const char *name, const uchar *mask, userflag uf)
   if (!name && !mask)		/* empty name has to have valid mask */
     return 0;
   rw_wrlock (&UFLock);
+  if (mask)
+  {
+    /* check to not add duplicate masks into Listfile */
+    user = _findbyhost (mask);
+    if (user)
+    {
+      ERROR ("Add_Clientrecord: hostmask pattern %s already exists.", mask);
+      rw_unlock (&UFLock);
+      return 0;
+    }
+  }
   user = _add_userrecord (name, uf, ID_ANY); /* client/ban */
   if (!user)
   {
@@ -668,8 +752,17 @@ int Add_Clientrecord (const char *name, const uchar *mask, userflag uf)
     return 0;
   }
   user->created = Time;
-  if (mask)			/* don't lock it, it's unreachable yet */
-    _add_usermask (user, mask);
+  if (mask &&			/* don't lock it, it's unreachable yet */
+      _add_usermask (user, mask) == 0)
+  {
+    if (!name)
+    {
+      ERROR ("Add_Clientrecord: invalid hostmask pattern %s.", mask);
+      _delete_userrecord (user, TRUE);
+      return 0;
+    }
+    ERROR ("Add_Clientrecord: invalid hostmask pattern %s ignored.", mask);
+  }
   rw_unlock (&UFLock);
   if (uf)
     Add_Request (I_LOG, "*", mask ? (F_USERS | F_AHEAD) : F_USERS,
@@ -835,6 +928,7 @@ static int _add_to_list2 (INTERFACE *iface, char *buf, size_t *len, char *msg)
   return n;
 }
 
+#define U_NONAMED (U_DENY | U_ACCESS | U_INVITE | U_DEOP | U_QUIET | U_IGNORED)
 /*--- W --- no locks ---*/
 int Get_Clientlist (INTERFACE *iface, userflag uf, const char *fn,
 		    const char *mask)
@@ -851,7 +945,7 @@ int Get_Clientlist (INTERFACE *iface, userflag uf, const char *fn,
 
   if (!mask || !*mask || !uf || !iface)
     return 0;
-  canbenonamed = uf & (U_DENY | U_ACCESS | U_INVITE | U_IGNORED);
+  canbenonamed = uf & U_NONAMED;
   fnisservice = safe_strchr ((char *)fn, '@');	/* isn't NULL if service name */
   gf = uf & U_GLOBALS;
   unistrlower (lcmask, mask, sizeof(lcmask));  
@@ -1046,7 +1140,10 @@ struct clrec_t *Find_Clientrecord (const uchar *mask, const char **lname,
   /* find the userrecord */
   pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cancelstate);
   rw_rdlock (&UFLock);
-  user = _findthebest (mask, NULL);
+  if (Have_Wildcard(mask) < 0)
+    user = _findthebest (mask, NULL);
+  else
+    user = _findbyhost (mask);
   if (user)
   {
     pthread_mutex_lock (&user->mutex);
@@ -1567,14 +1664,14 @@ int Add_Mask (struct clrec_t *user, const uchar *mask)
   int i = 0;
 
   if (!user || !mask) return 0;
-  if (!*mask)
+  if (*mask)
   {
-    ERROR ("Add_Mask: invalid hostmask pattern %s ignored.", mask);
-    return 0;
+    pthread_mutex_unlock (&user->mutex);	/* unlock to avoid conflicts */
+    i = _add_usermask (user, mask);
+    pthread_mutex_lock (&user->mutex);
   }
-  pthread_mutex_unlock (&user->mutex);		/* unlock to avoid conflicts */
-  i = _add_usermask (user, mask);
-  pthread_mutex_lock (&user->mutex);
+  if (i == 0)
+    ERROR ("Add_Mask: invalid hostmask pattern '%s' ignored.", mask);
   return i;
 }
 
@@ -1847,7 +1944,6 @@ static int _cral_clear (void)
   return r;
 }
 
-#define U_NONAMED (U_DENY | U_ACCESS | U_INVITE | U_DEOP | U_QUIET | U_IGNORED)
 /*
  * load mode:
  *	reset all
@@ -1898,7 +1994,8 @@ static int _load_listfile (const char *filename, int update)
     {
       case '+':				/* a hostmask */
 	if (ur)
-	  _add_usermask (ur, ++c);
+	  if (!_add_usermask (ur, ++c))
+	    ERROR ("_load_listfile: usermask %s ignored", c);
 	break;
       case 0:				/* empty line ignored */
       case '#':				/* an comment */
@@ -2761,12 +2858,12 @@ static int dc__phost (struct peer_t *dcc, char *args)
   if (!user)				/* no user */
     return 0;
   args = NextWord (args);
-  if (strlen (args) < 5 || match ("*.*", args) < 0) /* validation of hostmask */
+  if (strlen (args) < 5 || match ("*.*", args) < 0 ||
+      _add_usermask (user, args) == 0) /* validation of hostmask */
   {
     New_Request (dcc->iface, 0, "Invalid hostmask pattern: %s", args);
     return 0;
   }
-  _add_usermask (user, args);
   if (!(user->flag & (U_SPECIAL | U_UNSHARED)))	/* specials are not shared */
     Add_Request (I_DIRECT, "@*", F_SHARE, "\010+host %s %s", user->lname, args);
   return 1;
