@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2004-2011  Andrej N. Gritsenko <andrej@rep.kiev.ua>
+ * Copyright (C) 2004-2016  Andrej N. Gritsenko <andrej@rep.kiev.ua>
  *
  *     This program is free software; you can redistribute it and/or modify
  *     it under the terms of the GNU General Public License as published by
@@ -380,9 +380,7 @@ static int _irc_try_server (irc_server *serv, const char *tohost, int banned,
   irc_await *await;
 
   /* if already connected then break connection first */
-  if (Connchain_Kill ((&serv->p)) &&	/* condition to avoid warn */
-      serv->p.socket >= 0)
-    KillSocket(&serv->p.socket);
+  Peer_Cleanup (&serv->p);
   FREE (&serv->p.dname);
   serv->p.socket = -1;		/* it may still contain an error code */
   if (serv->p.state == P_TALK || serv->p.state == P_IDLE ||
@@ -824,7 +822,9 @@ static int _irc_request_main (INTERFACE *iface, REQUEST *req)
 		return _irc_try_server (serv, NULL, 0, NULL); /* abort it */
 	      }
 	Connchain_Grow ((&serv->p), 'x');	/* RAW -> TXT filter */
-	_irc_send (serv, thebuf); /* no errors should be here yet */
+	if (_irc_send (serv, thebuf) == 0)
+	  /* some filters may need handshake so queue a request for later */
+	  New_Request (iface, F_QUICK, "%s", thebuf);
       }
       else if (serv->p.socket < 0 || sw != E_AGAIN)
       {
@@ -873,7 +873,7 @@ static int _irc_request_main (INTERFACE *iface, REQUEST *req)
       serv->pmsgout = NULL;
 		/* "Connection reset by peer" */
 		/* note: I wonder if terminator might not send QUIT here yet */
-      KillSocket (&serv->p.socket);
+      Peer_Cleanup (&serv->p);
       if (serv->next)
 	serv->next->prev = serv->prev;
       if (serv->prev)
