@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2010-2011  Andrej N. Gritsenko <andrej@rep.kiev.ua>
+ * Copyright (C) 2010-2016  Andrej N. Gritsenko <andrej@rep.kiev.ua>
  *
  *     This program is free software; you can redistribute it and/or modify
  *     it under the terms of the GNU General Public License as published by
@@ -28,6 +28,7 @@
 #include "ircd.h"
 #include "numerics.h"
 
+extern bool _ircd_public_topic; /* in ircd.c */
 
 /* ---------------------------------------------------------------------------
  * Common internal functions.
@@ -105,8 +106,9 @@ static int ichmsg_ircd (modeflag umode, modeflag mmode, char *msg)
     return ircd_do_unumeric (b, ERR_NOTONCHANNEL, b, 0, c)
 
 BINDING_TYPE_ircd_client_cmd(ircd_oper_cb);
-static int ircd_oper_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick, char *user,
-			char *host, int argc, const char **argv)
+static int ircd_oper_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnick,
+			const char *user, const char *host, const char *vhost,
+			modeflag eum, int argc, const char **argv)
 { /* args: <name> <password> */
   CLIENT *cl = ((struct peer_priv *)peer->iface->data)->link->cl;
   struct clrec_t *u;
@@ -118,13 +120,16 @@ static int ircd_oper_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick, char
   int i;
 
   if (argc < 2)
-    return ircd_do_unumeric (cl, ERR_NEEDMOREPARAMS, cl, 0, NULL);
+    return ircd_do_unumeric (cl, ERR_NEEDMOREPARAMS, cl, 0, "OPER");
   u = Lock_Clientrecord (argv[0]);
   if (!u)
     return ircd_do_unumeric (cl, ERR_NOOPERHOST, cl, 0, NULL);
   fl = Get_Flags (u, srv->name);
   if (!(fl & (U_OP | U_HALFOP)))
+  {
+    Unlock_Clientrecord (u);
     return ircd_do_unumeric (cl, ERR_NOOPERHOST, cl, 0, NULL);
+  }
   /* we hope we never get OPER flood so doing memory allocation here */
   pass = safe_strdup (Get_Field (u, "passwd", NULL));
   id = Get_LID (u);
@@ -160,20 +165,36 @@ static int ircd_oper_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick, char
     New_Request (peer->iface, 0, ":%s MODE %s +o", peer->dname, peer->dname);
     ircd_sendto_servers_new ((IRCD *)srv->data, NULL, ":%s IMODE %d %s +o",
 			     peer->dname, ircd_new_id(), peer->dname);
+#ifdef USE_SERVICES
+    ircd_sendto_services_mark_nick ((IRCD *)srv->data, SERVICE_WANT_OPER);
+#endif
     ircd_sendto_servers_old ((IRCD *)srv->data, NULL, ":%s MODE %s +o",
 			     peer->dname, peer->dname);
+#ifdef USE_SERVICES
+    ircd_sendto_services_prefix ((IRCD *)srv->data, SERVICE_WANT_OPER,
+				 ":%s!%s@%s MODE %s +o", peer->dname, cl->user,
+				 cl->host, peer->dname);
+#endif
   }
   else
   {
     cl->umode |= A_HALFOP;		/* local oper */
     New_Request (peer->iface, 0, ":%s MODE %s +O", peer->dname, peer->dname);
+#ifdef USE_SERVICES
+    ircd_sendto_services_nick ((IRCD *)srv->data, SERVICE_WANT_OPER,
+			       ":%s MODE %s +O", peer->dname, peer->dname);
+    ircd_sendto_services_prefix ((IRCD *)srv->data, SERVICE_WANT_OPER,
+				 ":%s!%s@%s MODE %s +O", peer->dname, cl->user,
+				 cl->host, peer->dname);
+#endif
   }
   return ircd_do_unumeric (cl, RPL_YOUREOPER, cl, 0, NULL);
 }
 
 BINDING_TYPE_ircd_client_cmd(ircd_quit_cb);
-static int ircd_quit_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick, char *user,
-			char *host, int argc, const char **argv)
+static int ircd_quit_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnick,
+			const char *user, const char *host, const char *vhost,
+			modeflag eum, int argc, const char **argv)
 { /* args: [<Quit Message>] */
   CLIENT *cl = ((struct peer_priv *)peer->iface->data)->link->cl;
   char msg[STRING];
@@ -186,21 +207,27 @@ static int ircd_quit_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick, char
     snprintf (msg, sizeof(msg), "\"%.*s\"", (int)sz, argv[0]); /* quote it */
   }
 #ifdef USE_SERVICES
-  //TODO: do with services
+  ircd_sendto_services_mark_nick ((IRCD *)srv->data,
+				  SERVICE_WANT_QUIT | SERVICE_WANT_RQUIT);
 #endif
   ircd_sendto_servers_all_ack ((IRCD *)srv->data, cl, NULL, NULL,
 			       ":%s QUIT :%s", peer->dname, msg);
   ircd_prepare_quit (cl, cl->via, msg);
-  Add_Request (I_PENDING, "*", 0, ":%s!%s@%s QUIT :%s", peer->dname, cl->user,
-	       cl->host, msg);
+#ifdef USE_SERVICES
+  ircd_sendto_services_mark_prefix ((IRCD *)srv->data,
+				    SERVICE_WANT_QUIT | SERVICE_WANT_RQUIT);
+#endif
+  Add_Request (I_PENDING, "*", 0, ":%s!%s@%s QUIT :%s", peer->dname, user,
+	       vhost, msg);
   cl->hold_upto = Time;
   cl->host[0] = '\0';			/* for collision check */
   return 1;
 }
 
 BINDING_TYPE_ircd_client_cmd(ircd_squit_cb);
-static int ircd_squit_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick, char *user,
-			 char *host, int argc, const char **argv)
+static int ircd_squit_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnick,
+			 const char *user, const char *host, const char *vhost,
+			 modeflag eum, int argc, const char **argv)
 { /* args: <server> <comment> */
   CLIENT *cl = ((struct peer_priv *)peer->iface->data)->link->cl;
   CLIENT *tgt;
@@ -209,7 +236,7 @@ static int ircd_squit_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick, char
   {
     if (!(cl->umode & (A_OP | A_HALFOP)))
       ircd_do_unumeric (cl, ERR_NOPRIVILEGES, cl, 0, NULL);
-    return ircd_do_unumeric (cl, ERR_NEEDMOREPARAMS, cl, 0, NULL);
+    return ircd_do_unumeric (cl, ERR_NEEDMOREPARAMS, cl, 0, "SQUIT");
   }
   tgt = ircd_find_client (argv[0], NULL);
   if (!(cl->umode & (A_OP | A_HALFOP)))
@@ -231,15 +258,16 @@ static int ircd_squit_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick, char
 }
 
 BINDING_TYPE_ircd_client_cmd(ircd_part_cb);
-static int ircd_part_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick, char *user,
-			char *host, int argc, const char **argv)
+static int ircd_part_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnick,
+			const char *user, const char *host, const char *vhost,
+			modeflag eum, int argc, const char **argv)
 { /* args: <channel>[,<channel> ...] [<Part Message>] */
   CLIENT *cl = ((struct peer_priv *)peer->iface->data)->link->cl;
   char *c;
   const char *msg;
 
   if (argc == 0)
-    return ircd_do_unumeric (cl, ERR_NEEDMOREPARAMS, cl, 0, NULL);
+    return ircd_do_unumeric (cl, ERR_NEEDMOREPARAMS, cl, 0, "PART");
   if (argc == 1)
     msg = peer->dname;
   else
@@ -261,24 +289,21 @@ static int ircd_part_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick, char
     {
       if ((memb->chan->mode & A_QUIET)) /* notify only sender */
 	New_Request (peer->iface, 0, ":%s!%s@%s PART %s :%s", cl->nick, user,
-		     host, memb->chan->name, msg);
+		     vhost, memb->chan->name, msg);
       else				/* notify local users */
       {
 	if (memb->chan->mode & A_ANONYMOUS)
 	{
 	  New_Request (peer->iface, 0, ":%s!%s@%s PART %s :%s", cl->nick, user,
-		       host, memb->chan->name, msg);
+		       vhost, memb->chan->name, msg);
 	  ircd_sendto_chan_butone (memb->chan, cl,
 				   ":anonymous!anonymous@anonymous. PART %s :anonymous",
 				   memb->chan->name);
 	}
 	else
 	  ircd_sendto_chan_local (memb->chan, ":%s!%s@%s PART %s :%s", cl->nick,
-				  user, host, memb->chan->name, msg);
+				  user, vhost, memb->chan->name, msg);
       }
-#ifdef USE_SERVICES
-      //TODO: notify services
-#endif
       if (memb->chan->mode & A_INVISIBLE) ; /* local channel */
       else if ((cmask = strchr (c, ':'))) /* notify servers */
       {
@@ -299,8 +324,9 @@ static int ircd_part_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick, char
 }
 
 BINDING_TYPE_ircd_client_cmd(ircd_topic_cb);
-static int ircd_topic_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick, char *user,
-			 char *host, int argc, const char **argv)
+static int ircd_topic_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnick,
+			 const char *user, const char *host, const char *vhost,
+			 modeflag eum, int argc, const char **argv)
 { /* args: <channel> [<topic>] */
   CLIENT *cl = ((struct peer_priv *)peer->iface->data)->link->cl;
   MEMBER *memb;
@@ -309,30 +335,42 @@ static int ircd_topic_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick, char
   char *cmask;
 
   if (argc == 0)
-    return ircd_do_unumeric (cl, ERR_NEEDMOREPARAMS, cl, 0, NULL);
+    return ircd_do_unumeric (cl, ERR_NEEDMOREPARAMS, cl, 0, "TOPIC");
   memb = ircd_find_member ((IRCD *)srv->data, argv[0], NULL);
   if (memb == NOSUCHCHANNEL)
     return ircd_do_unumeric (cl, ERR_NOTONCHANNEL, cl, 0, argv[0]);
   ch = memb->chan;
   memb = _ircd_is_on_channel(cl, ch);
-#ifdef IRCD_PUBLIC_TOPIC
   /* secret channels should be not visible such way - RFC2811 */
-  if ((ch->mode & A_SECRET) && memb == NULL)
-#else
-  if (memb == NULL)
-#endif
+  if ((!_ircd_public_topic || (ch->mode & A_SECRET)) && memb == NULL)
     return ircd_do_unumeric (cl, ERR_NOTONCHANNEL, cl, 0, argv[0]);
   if (argc == 1)			/* it's query */
   {
     if (ch->topic[0])
-      return ircd_do_cnumeric (cl, RPL_TOPIC, ch, 0, ch->topic);
+    {
+      ircd_do_cnumeric (cl, RPL_TOPIC, ch, 0, ch->topic);
+#ifdef TOPICWHOTIME
+      if (ch->topic_since > 0)
+      {
+	char topicwhotime[HOSTMASKLEN+12]; /* nick time */
+
+	if (ch->mode & A_ANONYMOUS)
+	  snprintf (topicwhotime, sizeof(topicwhotime),
+		    "anonymous!anonymous@anonymous. %ld", ch->topic_since);
+	else
+	  snprintf (topicwhotime, sizeof(topicwhotime), "%s %ld", ch->topic_by,
+		    ch->topic_since);
+	ircd_do_cnumeric (cl, RPL_TOPICWHOTIME, ch, 0, topicwhotime);
+      }
+#endif
+      return 1;
+    }
     return ircd_do_cnumeric (cl, RPL_NOTOPIC, ch, 0, NULL);
   }
-#ifdef IRCD_PUBLIC_TOPIC
   if (memb == NULL)
     return ircd_do_unumeric (cl, ERR_NOTONCHANNEL, cl, 0, argv[0]);
-#endif
-  if ((ch->mode & A_TOPICLOCK) && !(memb->mode & (A_ADMIN | A_OP)))
+  if ((ch->mode & A_TOPICLOCK) && !(memb->mode & (A_ADMIN | A_OP)) &&
+      !(eum & A_SERVER)) /* override member mode for _ircd_do_command() */
   {
     if (ch->name[0] == '+')
       return ircd_do_cnumeric (cl, ERR_NOCHANMODES, ch, 0, NULL);
@@ -340,11 +378,19 @@ static int ircd_topic_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick, char
   }
   sz = unistrcut (argv[1], sizeof(ch->topic), TOPICLEN); /* validate */
   strfcpy (ch->topic, argv[1], sz+1);
+#ifdef TOPICWHOTIME
+  snprintf (ch->topic_by, sizeof(ch->topic_by), "%s!%s@%s", peer->dname, user,
+	    vhost);
+  ch->topic_since = Time;
+#endif
+#ifdef USE_SERVICES
+  ircd_sendto_services_mark_prefix ((IRCD *)srv->data, SERVICE_WANT_TOPIC);
+#endif
   if (ch->mode & A_ANONYMOUS)
     ircd_sendto_chan_local(ch, ":anonymous!anonymous@anonymous. TOPIC %s :%s",
 			   ch->name, ch->topic);
   else
-    ircd_sendto_chan_local(ch, ":%s!%s@%s TOPIC %s :%s", peer->dname, user, host,
+    ircd_sendto_chan_local(ch, ":%s!%s@%s TOPIC %s :%s", peer->dname, user, vhost,
 			   ch->name, ch->topic);
   if (ch->mode & A_INVISIBLE)		/* it's local channel */
     return 1;
@@ -352,6 +398,9 @@ static int ircd_topic_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick, char
   if (cmask)
   {
     cmask++; /* don't place this in macro below */
+#ifdef USE_SERVICES
+    ircd_sendto_services_mark_nick ((IRCD *)srv->data, SERVICE_WANT_TOPIC);
+#endif
     ircd_sendto_servers_mask_new ((IRCD *)srv->data, NULL, cmask,
 				  ":%s ITOPIC %d %s :%s", peer->dname,
 				  ircd_new_id(), ch->name, ch->topic);
@@ -360,22 +409,26 @@ static int ircd_topic_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick, char
 				  ch->topic);
     return 1;
   }
-  ircd_sendto_servers_new ((IRCD *)srv->data, NULL, ":%s ITOPIC %d %s :%s",
-			   peer->dname, ircd_new_id(), ch->name, ch->topic);
+#ifdef USE_SERVICES
+  ircd_sendto_services_mark_nick ((IRCD *)srv->data, SERVICE_WANT_TOPIC);
+#endif
   ircd_sendto_servers_old ((IRCD *)srv->data, NULL, ":%s TOPIC %s :%s",
 			   peer->dname, ch->name, ch->topic);
+  ircd_sendto_servers_new ((IRCD *)srv->data, NULL, ":%s ITOPIC %d %s :%s",
+			   peer->dname, ircd_new_id(), ch->name, ch->topic);
   return 1;
 }
 
 BINDING_TYPE_ircd_client_cmd(ircd_invite_cb);
-static int ircd_invite_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick, char *user,
-			  char *host, int argc, const char **argv)
+static int ircd_invite_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnick,
+			  const char *user, const char *host, const char *vhost,
+			  modeflag eum, int argc, const char **argv)
 { /* args: <nickname> <channel> */
   CLIENT *cl = ((struct peer_priv *)peer->iface->data)->link->cl, *tgt;
   MEMBER *memb;
 
   if (argc < 2)
-    return ircd_do_unumeric (cl, ERR_NEEDMOREPARAMS, cl, 0, NULL);
+    return ircd_do_unumeric (cl, ERR_NEEDMOREPARAMS, cl, 0, "INVITE");
   tgt = ircd_find_client (argv[0], NULL);
   if (!tgt || (tgt->umode & (A_SERVER|A_SERVICE)))
     return ircd_do_unumeric (cl, ERR_NOSUCHNICK, cl, 0, argv[0]);
@@ -395,7 +448,7 @@ static int ircd_invite_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick, cha
       return (1); /* TODO: send some numeric to client? */
     ircd_sendto_one (tgt, ":%s INVITE %s %s", peer->dname, argv[0], argv[1]);
   } else
-    ircd_sendto_one (tgt, ":%s!%s@%s INVITE %s %s", peer->dname, user, host,
+    ircd_sendto_one (tgt, ":%s!%s@%s INVITE %s %s", peer->dname, user, vhost,
 		     argv[0], argv[1]);
   if (!CLIENT_IS_REMOTE(tgt) && memb != NOSUCHCHANNEL)
     ircd_add_invited (tgt, memb->chan);
@@ -405,8 +458,9 @@ static int ircd_invite_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick, cha
 }
 
 BINDING_TYPE_ircd_client_cmd(ircd_kick_cb);
-static int ircd_kick_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick, char *user,
-			char *host, int argc, const char **argv)
+static int ircd_kick_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnick,
+			const char *user, const char *host, const char *vhost,
+			modeflag eum, int argc, const char **argv)
 { /* args: <channel>[,<channel> ...] <user>[,<user> ...] [<comment>] */
   CLIENT *cl = ((struct peer_priv *)peer->iface->data)->link->cl, *tgt;
   MEMBER *memb, *tm;
@@ -415,7 +469,7 @@ static int ircd_kick_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick, char
   register char *cmask;
 
   if (argc < 2)
-    return ircd_do_unumeric (cl, ERR_NEEDMOREPARAMS, cl, 0, NULL);
+    return ircd_do_unumeric (cl, ERR_NEEDMOREPARAMS, cl, 0, "KICK");
   lch = strchr (argv[0], ',');
   for (chn = nchn = (char *)argv[0], lcl = (char *)argv[1]; lcl;
 	lcl = nlcl, chn = nchn)
@@ -432,7 +486,8 @@ static int ircd_kick_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick, char
       ircd_do_unumeric (cl, ERR_NOSUCHCHANNEL, cl, 0, chn);
     else if (memb == NULL)
       ircd_do_unumeric (cl, ERR_NOTONCHANNEL, cl, 0, chn);
-    else if (!(memb->mode & (A_OP | A_ADMIN)))
+    else if (!(memb->mode & (A_OP | A_ADMIN)) && !(eum & A_SERVER))
+      /* override member mode if that was _ircd_do_command() */
       ircd_do_cnumeric (cl, ERR_CHANOPRIVSNEEDED, memb->chan, 0, NULL);
     else if (!(tgt = ircd_find_client (lcl, NULL)) ||
 	     !(tm = _ircd_is_on_channel (tgt, memb->chan)))
@@ -447,7 +502,7 @@ static int ircd_kick_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick, char
 	reason = peer->dname;
       if (memb->chan->mode & A_ANONYMOUS) {
 //	New_Request(cl->via->p.iface, 0, ":%s!%s@%s KICK %s %s :%s",
-//		    peer->dname, user, host, chn, lcl, reason);
+//		    peer->dname, user, vhost, chn, lcl, reason);
 	if (!CLIENT_IS_REMOTE(tgt))
 	  New_Request(tgt->via->p.iface, 0,
 		      ":anonymous!anonymous@anonymous. KICK %s %s :%s",
@@ -457,10 +512,7 @@ static int ircd_kick_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick, char
 				chn, reason);
       } else
 	ircd_sendto_chan_local (memb->chan, ":%s!%s@%s KICK %s %s :%s",
-				peer->dname, user, host, chn, lcl, reason);
-#ifdef USE_SERVICES
-      //TODO: inform services
-#endif
+				peer->dname, user, vhost, chn, lcl, reason);
       if (memb->chan->mode & A_INVISIBLE) ;
       else if ((cmask = strchr (memb->chan->name, ':'))) {
 	cmask++; /* not put '++' into macro below */
@@ -479,8 +531,9 @@ static int ircd_kick_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick, char
 }
 
 BINDING_TYPE_ircd_client_cmd(ircd_servlist_cb);
-static int ircd_servlist_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick, char *user,
-			    char *host, int argc, const char **argv)
+static int ircd_servlist_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnick,
+			    const char *user, const char *host, const char *vhost,
+			    modeflag eum, int argc, const char **argv)
 { /* args: [<mask>[ <type>]] */
   CLIENT *cl = ((struct peer_priv *)peer->iface->data)->link->cl;
   CLIENT *tgt;
@@ -505,7 +558,7 @@ static int ircd_servlist_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick, c
     {
       snprintf (buf, sizeof(buf), "%s %s %s %hu :%s",
 #ifdef USE_SERVICES
-      !CLIENT_IS_REMOTE (tgt) ? me :
+      !CLIENT_IS_REMOTE (tgt) ? me->nick :
 #endif
 		tgt->cs->nick, mask, tgt->away, tgt->hops - 1, tgt->fname);
       ircd_do_unumeric (cl, RPL_SERVLIST, tgt, 0, buf);
@@ -530,21 +583,22 @@ static inline void _ircd_who_reply (CLIENT *rq, CLIENT *srv, CLIENT *tgt,
     else
       ch[0] = ircd_mode2whochar (m->mode);
     snprintf (buf, sizeof(buf), "%s %s %s %s %s %c%s%s :%hu %s", m->chan->name,
-	      tgt->user, tgt->host, srv->lcnick, tgt->nick,
+	      tgt->user, tgt->vhost, srv->lcnick, tgt->nick,
 	      (tgt->umode & A_AWAY) ? 'G' : 'H',
 	      (tgt->umode & (A_OP | A_HALFOP)) ? "*" : "", ch, tgt->hops - 1,
 	      tgt->fname);
   }
   else
     snprintf (buf, sizeof(buf), "* %s %s %s %s %c%s :%hu %s", tgt->user,
-	      tgt->host, srv->lcnick, tgt->nick, (tgt->umode & A_AWAY) ? 'G' : 'H',
+	      tgt->vhost, srv->lcnick, tgt->nick, (tgt->umode & A_AWAY) ? 'G' : 'H',
 	      (tgt->umode & (A_OP | A_HALFOP)) ? "*" : "", tgt->hops - 1, tgt->fname);
   ircd_do_unumeric (rq, RPL_WHOREPLY, rq, 0, buf);
 }
 
 BINDING_TYPE_ircd_client_cmd(ircd_who_cb);
-static int ircd_who_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
-		       char *user, char *host, int argc, const char **argv)
+static int ircd_who_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnick,
+		       const char *user, const char *host, const char *vhost,
+		       modeflag eum, int argc, const char **argv)
 { /* args: [<mask>[ "o"]] */
   CLIENT *cl = ((struct peer_priv *)peer->iface->data)->link->cl;
   const char *mask = NULL;
@@ -599,6 +653,7 @@ static int ircd_who_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
 		  break;
 	  if (mc)
 	    if (!mask || smatched >= 0 || simple_match (mask, tgt->host) >= 0 ||
+		((tgt->umode & A_MASKED) && simple_match (mask, tgt->vhost) >= 0) ||
 		simple_match (mask, tgt->lcnick) >= 0 ||
 		simple_match (mask, tgt->fname) >= 0) //TODO: LC search?
 	      _ircd_who_reply (cl, (i == 0) ? me : tgt->cs, tgt, NULL);
@@ -622,8 +677,9 @@ static int ircd_who_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
 
 #ifdef IRCD_ENABLE_KILL
 BINDING_TYPE_ircd_client_cmd(ircd_kill_cb);
-static int ircd_kill_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
-			char *user, char *host, int argc, const char **argv)
+static int ircd_kill_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnick,
+			const char *user, const char *host, const char *vhost,
+			modeflag eum, int argc, const char **argv)
 { /* args: <nickname> <comment> */
   CLIENT *cl = ((struct peer_priv *)peer->iface->data)->link->cl, *tcl;
   char reason[MB_LEN_MAX*TOPICLEN+HOSTMASKLEN];
@@ -631,7 +687,7 @@ static int ircd_kill_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
   register char *c;
 
   if (argc < 2)
-    return ircd_do_unumeric(cl, ERR_NEEDMOREPARAMS, cl, 0, NULL);
+    return ircd_do_unumeric(cl, ERR_NEEDMOREPARAMS, cl, 0, "KILL");
   tcl = ircd_find_client(argv[0], NULL);
   if (tcl == NULL)
     return ircd_do_unumeric(cl, ERR_NOSUCHNICK, cl, 0, argv[0]);
@@ -641,12 +697,18 @@ static int ircd_kill_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
       !(CLIENT_IS_LOCAL(tcl) && (cl->umode & A_HALFOP))) /* local - only local */
     return ircd_do_unumeric(cl, ERR_NOPRIVILEGES, cl, 0, NULL);
   len = unistrcut(argv[1], sizeof(reason), TOPICLEN);
-  snprintf(reason, sizeof(reason), "%s!%s (%.*s)", cl->host, cl->nick, len,
+  snprintf(reason, sizeof(reason), "%s!%s (%.*s)", cl->vhost, cl->nick, len,
 	   argv[1]);			/* make the message with reason */
   //TODO: implement LOCAL_KILL_ONLY option
   if (!CLIENT_IS_REMOTE(tcl))
     New_Request(tcl->via->p.iface, 0, ":%s KILL %s :%s", cl->nick, tcl->nick,
 		reason);		/* notify the victim */
+#ifdef USE_SERVICES
+  ircd_sendto_services_prefix((IRCD *)srv->data, SERVICE_WANT_KILL,
+			      ":%s!%s@%s KILL %s :%s", cl->nick, cl->user,
+			      cl->vhost, tcl->nick, reason);
+  ircd_sendto_services_mark_nick((IRCD *)srv->data, SERVICE_WANT_KILL);
+#endif
   ircd_sendto_servers_all_ack((IRCD *)srv->data, tcl, NULL, NULL,
 			      ":%s KILL %s :%s", cl->nick, tcl->nick, reason);
 				/* broadcast KILL */
@@ -654,7 +716,7 @@ static int ircd_kill_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
   tcl->hold_upto = Time + CHASETIMELIMIT; /* make 'nick delay' */
   for (c = NextWord(reason); c > reason && c[-1] != '!'; c--); /* find nick */
   Add_Request(I_PENDING, "*", 0, ":%s!%s@%s QUIT :Killed by %s", tcl->nick,
-	      tcl->user, tcl->host, c);
+	      tcl->user, tcl->vhost, c);
   tcl->host[0] = 0;		/* for collision check */
   Add_Request(I_LOG, "*", F_MODES, "KILL %s :%s", tcl->nick, reason);
   return (1);
@@ -662,12 +724,18 @@ static int ircd_kill_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
 #endif
 
 BINDING_TYPE_ircd_client_cmd(ircd_away_cb);
-static int ircd_away_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
-			char *user, char *host, int argc, const char **argv)
+static int ircd_away_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnick,
+			const char *user, const char *host, const char *vhost,
+			modeflag eum, int argc, const char **argv)
 { /* args: [<text>] */
   CLIENT *cl = ((struct peer_priv *)peer->iface->data)->link->cl;
   register size_t len;
 
+#ifdef USE_SERVICES
+  /* forbidden for services */
+  if (CLIENT_IS_SERVICE(cl))
+    return 0;
+#endif
   if (argc == 0 || *argv[0] == '\0') { /* unaway */
     cl->away[0] = '\0';
     cl->umode &= ~A_AWAY;
@@ -675,6 +743,12 @@ static int ircd_away_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
 			    peer->dname, ircd_new_id(), peer->dname);
     ircd_sendto_servers_old((IRCD *)srv->data, NULL, ":%s MODE %s :-a",
 			    peer->dname, peer->dname);
+#ifdef USE_SERVICES
+    ircd_sendto_services_prefix((IRCD *)srv->data, SERVICE_WANT_AWAY,
+			        ":%s!%s@%s AWAY", peer->dname, cl->user, cl->vhost);
+    ircd_sendto_services_nick((IRCD *)srv->data, SERVICE_WANT_AWAY,
+			      ":%s AWAY", peer->dname);
+#endif
     return ircd_do_unumeric(cl, RPL_UNAWAY, cl, 0, NULL);
   }
   len = unistrcut(argv[0], sizeof(cl->away), AWAYLEN);
@@ -684,13 +758,21 @@ static int ircd_away_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
 			  peer->dname, ircd_new_id(), peer->dname);
   ircd_sendto_servers_old((IRCD *)srv->data, NULL, ":%s MODE %s :+a",
 			  peer->dname, peer->dname);
+#ifdef USE_SERVICES
+  ircd_sendto_services_prefix((IRCD *)srv->data, SERVICE_WANT_AWAY,
+			      ":%s!%s@%s AWAY :%s", peer->dname, cl->user,
+			      cl->vhost, cl->away);
+  ircd_sendto_services_nick((IRCD *)srv->data, SERVICE_WANT_AWAY,
+			    ":%s AWAY :%s", peer->dname, cl->away);
+#endif
   return ircd_do_unumeric(cl, RPL_NOWAWAY, cl, 0, NULL);
 }
 
 #ifdef IRCD_ENABLE_REHASH
 BINDING_TYPE_ircd_client_cmd(ircd_rehash_cb);
-static int ircd_rehash_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
-			  char *user, char *host, int argc, const char **argv)
+static int ircd_rehash_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnick,
+			  const char *user, const char *host, const char *vhost,
+			  modeflag eum, int argc, const char **argv)
 { /* args: none */
   CLIENT *cl = ((struct peer_priv *)peer->iface->data)->link->cl;
   static char cmd[] = ".rehash";
@@ -705,8 +787,9 @@ static int ircd_rehash_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
 
 #ifdef IRCD_ENABLE_DIE
 BINDING_TYPE_ircd_client_cmd(ircd_die_cb);
-static int ircd_die_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
-		       char *user, char *host, int argc, const char **argv)
+static int ircd_die_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnick,
+		       const char *user, const char *host, const char *vhost,
+		       modeflag eum, int argc, const char **argv)
 { /* args: none */
   CLIENT *cl = ((struct peer_priv *)peer->iface->data)->link->cl;
   static char cmd[] = ".die";
@@ -721,8 +804,9 @@ static int ircd_die_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
 
 #ifdef IRCD_ENABLE_RESTART
 BINDING_TYPE_ircd_client_cmd(ircd_restart_cb);
-static int ircd_restart_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
-			   char *user, char *host, int argc, const char **argv)
+static int ircd_restart_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnick,
+			   const char *user, const char *host, const char *vhost,
+			   modeflag eum, int argc, const char **argv)
 { /* args: none */
   CLIENT *cl = ((struct peer_priv *)peer->iface->data)->link->cl;
   static char cmd[] = ".restart";
@@ -736,8 +820,9 @@ static int ircd_restart_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
 #endif
 
 BINDING_TYPE_ircd_client_cmd(ircd_userhost_cb);
-static int ircd_userhost_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
-			    char *user, char *host, int argc, const char **argv)
+static int ircd_userhost_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnick,
+			    const char *user, const char *host, const char *vhost,
+			    modeflag eum, int argc, const char **argv)
 { /* args: <nickname>[ ...] */
   CLIENT *cl = ((struct peer_priv *)peer->iface->data)->link->cl, *tgt;
   int i;
@@ -747,7 +832,7 @@ static int ircd_userhost_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
   char buf[IRCMSGLEN-HOSTLEN-NICKLEN-9]; /* :server 302 user :reply\r\n */
 
   if (argc == 0)
-    return ircd_do_unumeric(cl, ERR_NEEDMOREPARAMS, cl, 0, NULL);
+    return ircd_do_unumeric(cl, ERR_NEEDMOREPARAMS, cl, 0, "USERHOST");
   s = 0;
   for (i = 0; i < argc; i++) {
     for (c = argv[i]; *c; c = NextWord((char *)c)) /* it's still const */ {
@@ -759,7 +844,7 @@ static int ircd_userhost_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
       if (tgt == NULL || CLIENT_IS_SERVER(tgt))
 	continue;
       if ((s + strlen(tgt->nick) + strlen(tgt->user) +
-	  strlen(tgt->host)) >= (sizeof(buf) - 5)) {
+	  strlen(tgt->vhost)) >= (sizeof(buf) - 5)) {
 	ircd_do_unumeric(cl, RPL_USERHOST, cl, 0, buf);
 	s = 0;
       }
@@ -775,7 +860,7 @@ static int ircd_userhost_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
 	buf[s++] = '+';
       s += strfcpy(&buf[s], tgt->user, sizeof(buf) - s);
       buf[s++] = '@';
-      s += strfcpy(&buf[s], tgt->host, sizeof(buf) - s);
+      s += strfcpy(&buf[s], tgt->vhost, sizeof(buf) - s);
     }
   }
   if (s)
@@ -784,8 +869,9 @@ static int ircd_userhost_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
 }
 
 BINDING_TYPE_ircd_client_cmd(ircd_ison_cb);
-static int ircd_ison_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
-			char *user, char *host, int argc, const char **argv)
+static int ircd_ison_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnick,
+			const char *user, const char *host, const char *vhost,
+			modeflag eum, int argc, const char **argv)
 { /* args: <nickname>[ ...] */
   CLIENT *cl = ((struct peer_priv *)peer->iface->data)->link->cl, *tgt;
   int i;
@@ -795,7 +881,7 @@ static int ircd_ison_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
   char buf[IRCMSGLEN-HOSTLEN-NICKLEN-9]; /* :server 303 user :reply\r\n */
 
   if (argc == 0)
-    return ircd_do_unumeric(cl, ERR_NEEDMOREPARAMS, cl, 0, NULL);
+    return ircd_do_unumeric(cl, ERR_NEEDMOREPARAMS, cl, 0, "ISON");
   s = 0;
   buf[0] = '\0';
   for (i = 0; i < argc; i++) {
@@ -820,8 +906,9 @@ static int ircd_ison_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
 
 #if IRCD_USES_ICONV
 BINDING_TYPE_ircd_client_cmd(ircd_charset_cb);
-static int ircd_charset_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
-			   char *user, char *host, int argc, const char **argv)
+static int ircd_charset_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnick,
+			   const char *user, const char *host, const char *vhost,
+			   modeflag eum, int argc, const char **argv)
 { /* args: [<charset>] */
   CLIENT *cl = ((struct peer_priv *)peer->iface->data)->link->cl;
   struct conversion_t *conv;
