@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2010-2011  Andrej N. Gritsenko <andrej@rep.kiev.ua>
+ * Copyright (C) 2010-2017  Andrej N. Gritsenko <andrej@rep.kiev.ua>
  *
  *     This program is free software; you can redistribute it and/or modify
  *     it under the terms of the GNU General Public License as published by
@@ -18,6 +18,10 @@
  * This file is a part of FoxEye IRCd module.
  */
 
+/* these are used in structures below */
+#include <direct.h>
+#include <tree.h>
+
 	/* "modes" for internal usage */
 #define A_ISON		(1<<0)
 #define A_PINGED	(1<<3)
@@ -61,10 +65,6 @@
 # define IRCD_ID_HISTORY 65536
 #endif
 
-/* these are used in structures below */
-#include <direct.h>
-#include <tree.h>
-
 typedef struct CLASS CLASS;		/* internally defined in ircd.c */
 typedef struct CLIENT CLIENT;		/* defined below */
 typedef struct MEMBER MEMBER;
@@ -74,7 +74,7 @@ typedef struct MASK
   struct MASK *next;
 //  time_t since;
   char what[HOSTMASKLEN+1];
-//  char by[MB_LEN_MASK*NICKLEN+1];
+//  char by[HOSTMASKLEN+1];
 } MASK;
 
 typedef struct CHANNEL
@@ -91,6 +91,10 @@ typedef struct CHANNEL
   int on_ack; /* waiting for acks from servers (references counter) */
 #endif
   char fc[2];
+#ifdef TOPICWHOTIME
+  char topic_by[HOSTMASKLEN+1];
+  time_t topic_since;
+#endif
   char topic[MB_LEN_MAX*TOPICLEN+1];
   char key[KEYLEN+1];
   char name[MB_LEN_MAX*CHANNAMELEN+1];
@@ -166,9 +170,12 @@ typedef struct LINK
    - channel members - via list c.hannels
    - server clients - reference is cs
    *** as soon user gone offline it should be deleted from each of those */
+/* while registering, fields have another values: ->fname has input from
+   USER, ->lcnick is empty, ->away and ->vhost are filled by PASS input */
 struct CLIENT
 {
   CLIENT *pcl;				/* list in class */
+  struct peer_priv *local;		/* not NULL if it's local connect */
   struct peer_priv *via;		/* the way to this link */
 #if IRCD_MULTICONNECT
   struct peer_priv *alt;		/* second shortest (link instance) */
@@ -179,7 +186,10 @@ struct CLIENT
   union {
     CLASS *class;			/* user's class */
     CLIENT *rto;			/* 'renamed to' if nick is on hold */
-    unsigned short int token;		/* server's token, unique */
+    struct _CLIENT_x_a {
+      unsigned short int token;		/* server's token, unique */
+      unsigned short int uc;		/* users count, for statistics */
+    } a;
   } x;
   union {
     MEMBER *hannels;			/* user's channels list */
@@ -211,7 +221,7 @@ struct CLIENT
   char host[HOSTLEN+1];			/* host ; LC ; split server if on hold
 						       empty if hold by acks */
 					/* and it's used for service distrib. */
-  //char vhost[HOSTLEN+1];		/* TODO: "visible host" to show to users */
+  char vhost[HOSTLEN+1];		/* "visible host" to show to users */
 };
 
 /* channel's member data */
@@ -230,7 +240,9 @@ struct MEMBER
 typedef struct IRCD
 {
   INTERFACE *iface;	/* has network name */
+  INTERFACE *sub;	/* I_CLIENT with name @network to collect messages */
   NODE *clients;	/* clients list (name->CLIENT) */
+  unsigned int lu, gu;	/* max users count for statistics */
   NODE *channels;	/* channels list (name->CHANNEL) */
   CLASS *users;		/* local users lists */
   LINK *servers;	/* local servers list */
@@ -281,15 +293,50 @@ struct peer_priv
 
 /* test for ME : valid always */
 #define CLIENT_IS_ME(x) ((x)->cs == NULL)
-/* test for local clients/servers : crashes on ME, invalid on phantoms */
-#define CLIENT_IS_LOCAL(x) ((x)->cs->via->link->cl == x)
-/* test for remote clients : invalid on ME or phantoms */
+/* test for local clients/servers : valid always */
+#define CLIENT_IS_LOCAL(x) ((x)->local != NULL && (x)->via == (x)->local)
+/* test for remote clients : invalid on servers or phantoms */
 #define CLIENT_IS_REMOTE(x) ((x)->via == NULL)
 /* test for any server : valid always */
 #define CLIENT_IS_SERVER(x) ((x)->umode & A_SERVER)
 /* test for any service : valid always */
 #define CLIENT_IS_SERVICE(x) ((x)->umode & A_SERVICE)
 
+#ifdef USE_SERVICES
+#define SERVICE_WANT_SERVICE	0x00000001 /* other services signing on/off */
+#define SERVICE_WANT_OPER	0x00000002 /* operators, included in umodes too */
+#define SERVICE_WANT_UMODE	0x00000004 /* user modes, iow + local modes */
+#define SERVICE_WANT_AWAY	0x00000008 /* away isn't propaged anymore.. */
+#define SERVICE_WANT_KILL	0x00000010 /* KILLs */
+#define SERVICE_WANT_NICK	0x00000020 /* all NICKs (new user, change) */
+#define SERVICE_WANT_USER	0x00000040 /* USER signing on */
+#define SERVICE_WANT_QUIT	0x00000080 /* all QUITs (users signing off) */
+#define SERVICE_WANT_SERVER	0x00000100 /* servers signing on */
+#define SERVICE_WANT_WALLOP	0x00000200 /* wallops */
+#define SERVICE_WANT_SQUIT	0x00000400 /* servers signing off */
+#define SERVICE_WANT_RQUIT	0x00000800 /* regular user QUITs (these which are also sent between servers) */
+#define SERVICE_WANT_MODE	0x00001000 /* channel modes (not +ov) */
+#define SERVICE_WANT_CHANNEL	0x00002000 /* channel creations/destructions */
+#define SERVICE_WANT_VCHANNEL	0x00004000 /* channel joins/parts */
+#define SERVICE_WANT_TOPIC	0x00008000 /* channel topics */
+
+#define SERVICE_WANT_PREFIX	0x00010000 /* to receive n!u@h instead of n */
+#define SERVICE_WANT_TOKEN	0x00020000 /* use serv token instead of name */
+#define SERVICE_WANT_EXTNICK	0x00040000 /* user extended NICK syntax */
+
+#define SERVICE_WANT_ERRORS	0x01000000 /* &ERRORS */
+#define SERVICE_WANT_NOTICES	0x02000000 /* &NOTICES */
+#define SERVICE_WANT_LOCAL	0x04000000 /* &LOCAL */
+#define SERVICE_WANT_NUMERICS	0x08000000 /* &NUMERICS */
+
+#define SERVICE_WANT_USERLOG	0x10000000 /* FNAME_USERLOG */
+#define SERVICE_WANT_CONNLOG	0x20000000 /* FNAME_CONNLOG */
+
+#define SERVICE_IS_TRUSTED	0x80000000 /* cannot be set by services */
+
+#define SERVICE_FLAGS(x) (((uint32_t *)(x)->away)[1])
+#endif
+
 	/* ircd.c common functions */
 /* args: replyto, RPL_* / ERR_*, target, ping, message ; returns 1 */
 /* example:  ircd_do_numeric (clp, RPL_VERSION, NULL, 0, version_string); */
@@ -317,7 +364,7 @@ static inline unsigned short int client2token (CLIENT *cl)
 {
   if (cl != NULL && !cl->hold_upto &&
       (CLIENT_IS_SERVER(cl) || !CLIENT_IS_LOCAL(cl)))
-    return cl->cs->x.token;
+    return cl->cs->x.a.token;
   return 0;
 }
 /* should be called after correction and error message;
@@ -326,6 +373,9 @@ static inline unsigned short int client2token (CLIENT *cl)
 int ircd_recover_done (struct peer_priv *, const char *);
 /* marks local +w users to send WALLOPS, returns own server name */
 const char *ircd_mark_wallops(void);
+/* special search for a regular client */
+CLIENT *ircd_find_by_userhost(const char *nick, int ns, const char *user, int us,
+			      const char *host, int hs);
 
 #if IRCD_MULTICONNECT
 /* args: link, who, where */
@@ -370,16 +420,18 @@ char ircd_mode2whochar (modeflag);
 /* does backward conversion */
 modeflag ircd_whochar2mode(char);
 /* args: ircd, servername, nick, channame, mode char
+   used only when some remote user joins channel
    returns starting channel mode for '\0' or user-on-channel mode */
 modeflag ircd_char2mode(INTERFACE *, const char *, const char *, const char *, char);
-/* args: ircd, servername, mode char; returns umode for server modechange */
-modeflag ircd_char2umode(INTERFACE *, const char *, char);
+/* args: ircd, servername, mode char, user; returns umode for server modechange */
+modeflag ircd_char2umode(INTERFACE *, const char *, char, CLIENT *);
 
 	/* calls to channel.c from ircd.c */
 void ircd_channels_flush (IRCD *, char *, size_t);
 void ircd_burst_channels (INTERFACE *, NODE *);
 void ircd_channels_report (INTERFACE *);
 void ircd_channels_chreop (IRCD *, CLIENT *);
+void send_isupport (IRCD *, CLIENT *);
 
 	/* calls to client.c from channel.c */
 int ircd_names_reply (CLIENT *, CLIENT *, CHANNEL *, int);
@@ -406,4 +458,7 @@ void ircd_queries_register (void);
 void ircd_message_proto_start (void);
 void ircd_message_proto_end (void);
 
+	/* from queries.c */
+const char *ircd_version_flags;
+
 #include "sendto.h"
