@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2010-2011  Andrej N. Gritsenko <andrej@rep.kiev.ua>
+ * Copyright (C) 2010-2016  Andrej N. Gritsenko <andrej@rep.kiev.ua>
  *
  *     This program is free software; you can redistribute it and/or modify
  *     it under the terms of the GNU General Public License as published by
@@ -21,10 +21,14 @@
 #include <foxeye.h>
 #if IRCD_USES_ICONV == 0 || (defined(HAVE_ICONV) && (IRCD_NEEDS_TRANSLIT == 0 || defined(HAVE_CYRILLIC_TRANSLIT)))
 #include <modules.h>
+#include <init.h>
 
 #include "ircd.h"
 #include "numerics.h"
 
+extern bool _ircd_idle_from_msg; /* in ircd.c */
+extern short *_ircd_client_recvq;
+
 static struct bindtable_t *BTIrcdCheckMessage;
 static struct bindtable_t *BTIrcdSetMessageTargets;
 
@@ -251,7 +255,9 @@ static void _ircd_bmsgl_mask (IRCD *ircd, const char *t, char *mask,
     {
       tgt = l->s.data;
       if ((tgt->umode & (A_SERVER | A_SERVICE)) == m && !tgt->hold_upto &&
-	  !CLIENT_IS_REMOTE(tgt) && simple_match (mask, tgt->host) > 0)
+	  !CLIENT_IS_REMOTE(tgt) && (simple_match (mask, tgt->host) > 0 ||
+				     ((tgt->umode & A_MASKED) &&
+				      simple_match (mask, tgt->vhost) > 0)))
 	tgt->via->p.iface->ift |= I_PENDING;
     }
     if (!user)				/* service */
@@ -290,7 +296,8 @@ static int _ircd_check_server_clients_hosts (LINK *tgt, const char *mask)
       if (_ircd_check_server_clients_hosts (tgt, mask))
 	return 1;
     } else if (!(tgt->cl->umode & A_SERVICE) && !tgt->cl->hold_upto)
-      if (simple_match (mask, tgt->cl->host) > 0)
+      if (simple_match (mask, tgt->cl->host) > 0 ||
+	  ((tgt->cl->umode & A_MASKED) && simple_match (mask, tgt->cl->vhost) > 0))
 	return 1;
   return 0;
 }
@@ -303,19 +310,25 @@ static int _ircd_can_send_to_chan (CLIENT *cl, CHANNEL *ch, const char *msg)
   struct binding_t *b = NULL;
   int i, x = -1;
   char buff[MB_LEN_MAX*NICKLEN+IDENTLEN+HOSTLEN+3];
+  char buffv[MB_LEN_MAX*NICKLEN+IDENTLEN+HOSTLEN+3];
 
   m = _ircd_is_on_channel (cl, ch);
   if (m)
     mf = m->mode;
   snprintf (buff, sizeof(buff), "%s!%s@%s", cl->lcnick, cl->user, cl->host);
+  buffv[0] = '\0';
+  if (cl->umode & A_MASKED)
+    snprintf (buffv, sizeof(buffv), "%s!%s@%s", cl->lcnick, cl->user, cl->vhost);
   /* note: check all bans can be slow, I know, but what else to do? */
   for (cm = ch->bans; cm; cm = cm->next)
-    if (simple_match (cm->what, buff) > 0)
+    if (simple_match (cm->what, buff) > 0 ||
+	(buffv[0] && simple_match (cm->what, buffv) > 0))
       break;
   if (cm)
   {
     for (cm = ch->exempts; cm; cm = cm->next)
-      if (simple_match (cm->what, buff) > 0)
+      if (simple_match (cm->what, buff) > 0 ||
+	  (buffv[0] && simple_match (cm->what, buffv) > 0))
 	break;
     if (!cm)
       mf |= A_DENIED;
@@ -393,6 +406,7 @@ static void _ircd_broadcast_msglist_mark(IRCD *ircd, const char *nick,
       _ircd_mark_message_target(ircd->iface, nick, tlist[i]);
     } else
       tcl->cs->via->p.iface->ift |= I_PENDING;
+      //FIXME: do alternate way for exact targets?
   }
   if (need_unmark)
     for (lnk = ircd->token[0]->c.lients; lnk; lnk = lnk->prev) /* no locals */
@@ -410,13 +424,13 @@ static int _ircd_broadcast_msglist_new (IRCD *ircd, struct peer_priv *via,
 
   for (srv = ircd->servers; srv; srv = srv->prev) /* preset to ignore later */
     if (!(srv->cl->umode & A_MULTI) || srv->cl->via == via ||
-	srv->cl->x.token == token)
+	srv->cl->x.a.token == token)
       srv->cl->via->p.iface->ift |= I_PENDING;
   _ircd_broadcast_msglist_mark(ircd, nick, tlist, s);
   rc = 0;
   for (srv = ircd->servers; srv; srv = srv->prev) /* reset them now */
     if (!(srv->cl->umode & A_MULTI) || srv->cl->via == via ||
-	srv->cl->x.token == token)
+	srv->cl->x.a.token == token)
       srv->cl->via->p.iface->ift &= ~I_PENDING;
     else if (srv->cl->via->p.iface->ift & I_PENDING) /* it's targetted */
       rc = 1;
@@ -442,7 +456,7 @@ static int _ircd_broadcast_msglist_old (IRCD *ircd, struct peer_priv *via,
 #if IRCD_MULTICONNECT
 	(!all && (srv->cl->umode & A_MULTI)) ||
 #endif
-	srv->cl->x.token == token)
+	srv->cl->x.a.token == token)
       srv->cl->via->p.iface->ift |= I_PENDING;
   _ircd_broadcast_msglist_mark(ircd, nick, tlist, s);
   rc = 0;
@@ -451,7 +465,7 @@ static int _ircd_broadcast_msglist_old (IRCD *ircd, struct peer_priv *via,
 #if IRCD_MULTICONNECT
 	(!all && (srv->cl->umode & A_MULTI)) ||
 #endif
-	srv->cl->x.token == token)
+	srv->cl->x.a.token == token)
       srv->cl->via->p.iface->ift &= ~I_PENDING;
     else if (srv->cl->via->p.iface->ift & I_PENDING) /* it's targetted */
       rc = 1;
@@ -463,12 +477,39 @@ static int _ircd_broadcast_msglist_old (IRCD *ircd, struct peer_priv *via,
 static inline CLIENT *_ircd_find_msg_target (const char *target,
 					     struct peer_priv *pp)
 {
-  CLIENT *tgt = ircd_find_client (target, pp);
+  CLIENT *tgt;
+  const char *c, *h;
 
-  if (tgt && CLIENT_IS_SERVER(tgt))
+  /* handle nickname!user@host */
+  if ((c = strchr(target, '!')))
+  {
+    h = strchr(c, '@');
+    if (h)
+      tgt = ircd_find_by_userhost(target, c - target, c + 1, h - c - 1, h + 1, -1);
+    else
+      tgt = NULL;
+  }
+  /* handle user[%host]@servername */
+  else if ((c = strchr(target, '@')))
+  {
+    tgt = ircd_find_client (c + 1, pp);
+    if (tgt && !CLIENT_IS_SERVER(tgt))
+      tgt = NULL;
+    if (tgt && !CLIENT_IS_ME(tgt))
+      return (tgt);
+    /* process user[%host] now */
+    h = strchr(target, '%');
+    if (h > c)
+      h = NULL;
+    tgt = ircd_find_by_userhost(NULL, 0, target, h ? h - target : c - target,
+				h ? h + 1 : NULL, h ? c - h - 1 : 0);
+  }
+  /* handle user%host */
+  else if ((c = strchr(target, '%')))
+    tgt = ircd_find_by_userhost(NULL, 0, target, c - target, c + 1, -1);
+  else if ((tgt = ircd_find_client (target, pp)) && CLIENT_IS_SERVER(tgt))
     return (NULL);
   return (tgt);
-  //TODO: proc not plain too: user[%host]@servername user%host nickname!user@host
 }
 
 #if 0
@@ -496,7 +537,7 @@ static inline CLIENT *_ircd_find_msg_target (const char *target,
 #endif
 
 #define ADD_TO_LIST(S) \
-  if (s < MAXTARGETS) \
+  if (s < max_targets) \
     tlist[s++] = S; \
   if (s2 && s2 < sizeof(targets) - 2) /* reserved for ,x */ \
     targets[s2++] = ','; \
@@ -511,31 +552,37 @@ static inline CLIENT *_ircd_find_msg_target (const char *target,
    and any type of messages for it from servers should generate an error!!! */
 
 BINDING_TYPE_ircd_client_cmd(ircd_privmsg_cb);
-static int ircd_privmsg_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
-			   char *user, char *host, int argc, const char **argv)
+static int ircd_privmsg_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnick,
+			   const char *user, const char *host, const char *vhost,
+			   modeflag eum, int argc, const char **argv)
 { /* args: <msgtarget> <text to be sent> */
   CLIENT *cl = ((struct peer_priv *)peer->iface->data)->link->cl, *tcl;
   MEMBER *tch;
   IRCD *ircd = (IRCD *)srv->data;
   char *c, *cnext;
   register char *cmask;
-  int n;
-  const char *tlist[MAXTARGETS];
+  unsigned int n;
+  unsigned int max_targets = _ircd_client_recvq[0];
+  const char *tlist[max_targets];
   size_t s = 0, s2 = 0;
   char targets[MESSAGEMAX];
 
+#ifdef USE_SERVICES
+  /* forbidden for services */
+  if (CLIENT_IS_SERVICE(cl))
+    return 0;
+#endif
   if (argc == 0 || !*argv[0])
     return ircd_do_unumeric (cl, ERR_NORECIPIENT, cl, 0, NULL);
   if (argc == 1 || !*argv[1])
     return ircd_do_unumeric (cl, ERR_NOTEXTTOSEND, cl, 0, NULL);
-#ifdef IDLE_FROM_MSG
-  ((struct peer_priv *)peer->iface->data)->noidle = Time;
-#endif
+  if (_ircd_idle_from_msg)
+    ((struct peer_priv *)peer->iface->data)->noidle = Time;
   for (c = (char *)argv[0], n = 0; c; c = cnext)
   {
     if ((cnext = strchr (c, ',')))
       *cnext++ = 0;
-    if (++n > MAXTARGETS)
+    if (++n > max_targets)
       ircd_do_unumeric (cl, ERR_TOOMANYTARGETS, cl, 0, "No Message Delivered.");
     else if ((tch = ircd_find_member (ircd, c, NULL))
 	     != NOSUCHCHANNEL)
@@ -545,7 +592,7 @@ static int ircd_privmsg_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
 	ircd_do_unumeric (cl, ERR_NOSUCHNICK, cl, 0, c);
       else if (_ircd_can_send_to_chan (cl, tch->chan, argv[1]))
       {
-	_ircd_bmsgl_chan (tch->chan, cl, user, host, "PRIVMSG", argv[1]);
+	_ircd_bmsgl_chan (tch->chan, cl, user, vhost, "PRIVMSG", argv[1]);
 	if (!(tch->mode & A_INVISIBLE))
 	  ADD_TO_LIST(tch->chan->lcname);
       }
@@ -565,11 +612,11 @@ static int ircd_privmsg_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
 #ifdef USE_SERVICES
 	if (CLIENT_IS_SERVICE (cl))
 	  _ircd_bmsgl_mask (ircd, c, &c[1], 0, peer->dname, NULL,
-			    ircd_find_client(NULL, NULL), "PRIVMSG", argv[1]);
+			    ircd->token[0]->lcnick, "PRIVMSG", argv[1]);
 	else
 #endif
 	  _ircd_bmsgl_mask (ircd, c, &c[1], 0, peer->dname, user,
-			    host, "PRIVMSG", argv[1]);
+			    vhost, "PRIVMSG", argv[1]);
 	ADD_TO_LIST(c); //TODO: lowercase it?
       }
     }
@@ -582,11 +629,11 @@ static int ircd_privmsg_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
 #ifdef USE_SERVICES
 	if (CLIENT_IS_SERVICE (cl))
 	  New_Request (tcl->via->p.iface, 0, ":%s@%s PRIVMSG %s :%s",
-		       peer->dname, ircd_find_client(NULL, NULL), c, argv[1]);
+		       peer->dname, ircd->token[0]->lcnick, c, argv[1]);
 	else
 #endif
 	  New_Request (tcl->via->p.iface, 0, ":%s!%s@%s PRIVMSG %s :%s",
-		       peer->dname, user, host, c, argv[1]);
+		       peer->dname, user, vhost, c, argv[1]);
 	if (tcl->umode & A_AWAY)
 	  ircd_do_unumeric (cl, RPL_AWAY, tcl, 0, tcl->away);
       }
@@ -607,11 +654,11 @@ static int ircd_privmsg_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
 #ifdef USE_SERVICES
 	if (CLIENT_IS_SERVICE (cl))
 	  Add_Request(I_PENDING, "*", 0, ":%s@%s PRIVMSG %s :%s",
-		      peer->dname, ircd_find_client(NULL, NULL), c, argv[1]);
+		      peer->dname, ircd->token[0]->lcnick, c, argv[1]);
 	else
 #endif
 	  Add_Request(I_PENDING, "*", 0, ":%s!%s@%s PRIVMSG %s :%s",
-		      peer->dname, user, host, c, argv[1]);
+		      peer->dname, user, vhost, c, argv[1]);
 	ADD_TO_LIST(c);
       } else
 	ircd_do_unumeric (cl, ERR_NOSUCHNICK, cl, 0, c);
@@ -628,16 +675,18 @@ static int ircd_privmsg_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
 }
 
 BINDING_TYPE_ircd_client_cmd(ircd_notice_cb);
-static int ircd_notice_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
-			  char *user, char *host, int argc, const char **argv)
+static int ircd_notice_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnick,
+			  const char *user, const char *host, const char *vhost,
+			  modeflag eum, int argc, const char **argv)
 { /* args: <msgtarget> <text to be sent> */
   CLIENT *cl = ((struct peer_priv *)peer->iface->data)->link->cl, *tcl;
   MEMBER *tch;
   IRCD *ircd = (IRCD *)srv->data;
   char *c, *cnext;
   register char *cmask;
-  int n;
-  const char *tlist[MAXTARGETS];
+  unsigned int n;
+  unsigned int max_targets = _ircd_client_recvq[0];
+  const char *tlist[max_targets];
   size_t s = 0, s2 = 0;
   char targets[MESSAGEMAX];
 
@@ -649,13 +698,13 @@ static int ircd_notice_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
   {
     if ((cnext = strchr (c, ',')))
       *cnext++ = 0;
-    if (++n > MAXTARGETS)
+    if (++n > max_targets)
       continue;
     else if ((tch = ircd_find_member (ircd, c, NULL)) != NOSUCHCHANNEL)
     {
       if (_ircd_can_send_to_chan (cl, tch->chan, argv[1]))
       {
-	_ircd_bmsgl_chan (tch->chan, cl, user, host, "NOTICE", argv[1]);
+	_ircd_bmsgl_chan (tch->chan, cl, user, vhost, "NOTICE", argv[1]);
 	if (!(tch->mode & A_INVISIBLE))
 	  ADD_TO_LIST(tch->chan->lcname);
       }
@@ -672,11 +721,11 @@ static int ircd_notice_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
 #ifdef USE_SERVICES
 	if (CLIENT_IS_SERVICE (cl))
 	  _ircd_bmsgl_mask (ircd, c, &c[1], 0, peer->dname, NULL,
-			    ircd_find_client(NULL, NULL), "NOTICE", argv[1]);
+			    ircd->token[0]->lcnick, "NOTICE", argv[1]);
 	else
 #endif
 	  _ircd_bmsgl_mask (ircd, c, &c[1], 0, peer->dname, user,
-			    host, "NOTICE", argv[1]);
+			    vhost, "NOTICE", argv[1]);
 	ADD_TO_LIST(c); //TODO: lowercase it?
       }
     }
@@ -687,11 +736,11 @@ static int ircd_notice_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
 #ifdef USE_SERVICES
 	if (CLIENT_IS_SERVICE (cl))
 	  New_Request (tcl->via->p.iface, 0, ":%s@%s NOTICE %s :%s",
-		       peer->dname, ircd_find_client(NULL, NULL), c, argv[1]);
+		       peer->dname, ircd->token[0]->lcnick, c, argv[1]);
 	else
 #endif
 	  New_Request (tcl->via->p.iface, 0, ":%s!%s@%s NOTICE %s :%s",
-		       peer->dname, user, host, c, argv[1]);
+		       peer->dname, user, vhost, c, argv[1]);
       else
 	ADD_TO_LIST(tcl->lcnick);
     }
@@ -709,11 +758,11 @@ static int ircd_notice_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
 #ifdef USE_SERVICES
 	if (CLIENT_IS_SERVICE (cl))
 	  Add_Request(I_PENDING, "*", 0, ":%s@%s NOTICE %s :%s",
-		      peer->dname, ircd_find_client(NULL, NULL), c, argv[1]);
+		      peer->dname, ircd->token[0]->lcnick, c, argv[1]);
 	else
 #endif
 	  Add_Request(I_PENDING, "*", 0, ":%s!%s@%s NOTICE %s :%s",
-		      peer->dname, user, host, c, argv[1]);
+		      peer->dname, user, vhost, c, argv[1]);
 	ADD_TO_LIST(c);
       }
     }
@@ -729,8 +778,9 @@ static int ircd_notice_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
 }
 
 BINDING_TYPE_ircd_client_cmd(ircd_squery_cb);
-static int ircd_squery_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
-			  char *user, char *host, int argc, const char **argv)
+static int ircd_squery_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnick,
+			  const char *user, const char *host, const char *vhost,
+			  modeflag eum, int argc, const char **argv)
 { /* args: <servicename> <text to be sent> */
   CLIENT *cl = ((struct peer_priv *)peer->iface->data)->link->cl, *tcl;
   IRCD *ircd = (IRCD *)srv->data;
@@ -745,7 +795,7 @@ static int ircd_squery_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
 #ifdef USE_SERVICES
   if (!CLIENT_IS_REMOTE (tcl))
   {
-    New_Request (tcl->via->p.iface, 0, ":%s SQUERY %s :%s", peer->dname, c,
+    New_Request (tcl->via->p.iface, 0, ":%s SQUERY %s :%s", peer->dname, argv[0],
 		 argv[1]);
     return 1;
   }
@@ -765,7 +815,7 @@ static int ircd_squery_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
 /* not defining __TRANSIT__ here since token is handled by _ircd_broadcast_* */
 BINDING_TYPE_ircd_server_cmd(ircd_privmsg_sb);
 static int ircd_privmsg_sb(INTERFACE *srv, struct peer_t *peer, unsigned short token,
-			   const char *sender, const char *lcsender, char *cmd,
+			   const char *sender, const char *lcsender,
 			   int argc, const char **argv)
 { /* args: <msgtarget> <text to be sent> */
   CLIENT *cl, *tcl;
@@ -773,7 +823,8 @@ static int ircd_privmsg_sb(INTERFACE *srv, struct peer_t *peer, unsigned short t
   IRCD *ircd = (IRCD *)srv->data;
   char *c, *cnext;
   struct peer_priv *pp = peer->iface->data; /* it's really peer */
-  const char *tlist[MAXTARGETS];
+  unsigned int max_targets = _ircd_client_recvq[0];
+  const char *tlist[max_targets];
   size_t s = 0, s2 = 0;
   char targets[MESSAGEMAX];
 
@@ -788,7 +839,7 @@ static int ircd_privmsg_sb(INTERFACE *srv, struct peer_t *peer, unsigned short t
   {
     if ((cnext = strchr(c, ',')))
       *cnext++ = 0;
-    if (s == MAXTARGETS) {
+    if (s == max_targets) {
       _ircd_broadcast_msglist_old(ircd, pp, token, sender, 1,
 				  targets, tlist, s, "PRIVMSG", argv[1]);
       s = s2 = 0;
@@ -801,7 +852,7 @@ static int ircd_privmsg_sb(INTERFACE *srv, struct peer_t *peer, unsigned short t
 		    sender, peer->dname);
       if (!(tch->mode & A_INVISIBLE)) {
 	ADD_TO_LIST(tch->chan->lcname);
-	_ircd_bmsgl_chan(tch->chan, cl, cl->user, cl->host, "PRIVMSG", argv[1]);
+	_ircd_bmsgl_chan(tch->chan, cl, cl->user, cl->vhost, "PRIVMSG", argv[1]);
       } else {
 	ERROR("ircd:PRIVMSG via %s for channel %s ignored", peer->dname, c);
 	ircd_recover_done(pp, "Invalid message target");
@@ -812,7 +863,7 @@ static int ircd_privmsg_sb(INTERFACE *srv, struct peer_t *peer, unsigned short t
 			 cl->cs->lcnick, "PRIVMSG", argv[1]);
       else
 	_ircd_bmsgl_mask(ircd, c, &c[1], 0, sender, cl->user,
-			 cl->host, "PRIVMSG", argv[1]);
+			 cl->vhost, "PRIVMSG", argv[1]);
       ADD_TO_LIST(c); //TODO: lowercase it?
     } else if ((tcl = _ircd_find_msg_target(c, pp))) {
       if (CLIENT_IS_SERVICE(tcl)) {
@@ -824,7 +875,7 @@ static int ircd_privmsg_sb(INTERFACE *srv, struct peer_t *peer, unsigned short t
 		      cl->cs->lcnick, c, argv[1]);
 	else
 	  New_Request(tcl->via->p.iface, 0, ":%s!%s@%s PRIVMSG %s :%s", sender,
-		      cl->user, cl->host, c, argv[1]);
+		      cl->user, cl->vhost, c, argv[1]);
 	if (tcl->umode & A_AWAY)
 	  ircd_do_unumeric(cl, RPL_AWAY, tcl, 0, tcl->away);
       } else
@@ -844,7 +895,7 @@ static int ircd_privmsg_sb(INTERFACE *srv, struct peer_t *peer, unsigned short t
 		      cl->cs->lcnick, c, argv[1]);
 	else
 	  Add_Request(I_PENDING, "*", 0, ":%s!%s@%s PRIVMSG %s :%s", sender,
-		      cl->user, cl->host, c, argv[1]);
+		      cl->user, cl->vhost, c, argv[1]);
 	ADD_TO_LIST(c);
       } else {
 	ERROR("ircd:invalid PRIVMSG target %s via %s", c, peer->dname);
@@ -860,7 +911,7 @@ static int ircd_privmsg_sb(INTERFACE *srv, struct peer_t *peer, unsigned short t
 
 BINDING_TYPE_ircd_server_cmd(ircd_notice_sb);
 static int ircd_notice_sb(INTERFACE *srv, struct peer_t *peer, unsigned short token,
-			  const char *sender, const char *lcsender, char *cmd,
+			  const char *sender, const char *lcsender,
 			  int argc, const char **argv)
 { /* args: <msgtarget> <text to be sent> */
   CLIENT *cl, *tcl;
@@ -868,7 +919,8 @@ static int ircd_notice_sb(INTERFACE *srv, struct peer_t *peer, unsigned short to
   IRCD *ircd = (IRCD *)srv->data;
   char *c, *cnext;
   struct peer_priv *pp = peer->iface->data; /* it's really peer */
-  const char *tlist[MAXTARGETS];
+  unsigned int max_targets = _ircd_client_recvq[0];
+  const char *tlist[max_targets];
   size_t s = 0, s2 = 0;
   char targets[MESSAGEMAX];
 
@@ -884,7 +936,7 @@ static int ircd_notice_sb(INTERFACE *srv, struct peer_t *peer, unsigned short to
   {
     if ((cnext = strchr(c, ',')))
       *cnext++ = 0;
-    if (s == MAXTARGETS) {
+    if (s == max_targets) {
       _ircd_broadcast_msglist_old(ircd, pp, token, sender, 1,
 				  targets, tlist, s, "NOTICE", argv[1]);
       s = s2 = 0;
@@ -897,7 +949,7 @@ static int ircd_notice_sb(INTERFACE *srv, struct peer_t *peer, unsigned short to
 		    sender, peer->dname);
       if (!(tch->mode & A_INVISIBLE)) {
 	ADD_TO_LIST(tch->chan->lcname);
-	_ircd_bmsgl_chan(tch->chan, cl, cl->user, cl->host, "NOTICE", argv[1]);
+	_ircd_bmsgl_chan(tch->chan, cl, cl->user, cl->vhost, "NOTICE", argv[1]);
       } else
 	Add_Request(I_LOG, "*", F_WARN, "ircd:invalid NOTICE target %s via %s",
 		    c, peer->dname);
@@ -907,7 +959,7 @@ static int ircd_notice_sb(INTERFACE *srv, struct peer_t *peer, unsigned short to
 			 cl->cs->lcnick, "NOTICE", argv[1]);
       else
 	_ircd_bmsgl_mask(ircd, c, &c[1], 0, sender, cl->user,
-			 cl->host, "NOTICE", argv[1]);
+			 cl->vhost, "NOTICE", argv[1]);
       ADD_TO_LIST(c); //TODO: lowercase it?
     } else if ((tcl = _ircd_find_msg_target(c, pp))) {
       if (CLIENT_IS_SERVICE(tcl)) {
@@ -919,7 +971,7 @@ static int ircd_notice_sb(INTERFACE *srv, struct peer_t *peer, unsigned short to
 		      cl->cs->lcnick, c, argv[1]);
 	else
 	  New_Request(tcl->via->p.iface, 0, ":%s!%s@%s NOTICE %s :%s", sender,
-		      cl->user, cl->host, c, argv[1]);
+		      cl->user, cl->vhost, c, argv[1]);
       } else
 	ADD_TO_LIST(tcl->lcnick);
     } else {
@@ -937,7 +989,7 @@ static int ircd_notice_sb(INTERFACE *srv, struct peer_t *peer, unsigned short to
 		      cl->cs->lcnick, c, argv[1]);
 	else
 	  Add_Request(I_PENDING, "*", 0, ":%s!%s@%s NOTICE %s :%s", sender,
-		      cl->user, cl->host, c, argv[1]);
+		      cl->user, cl->vhost, c, argv[1]);
 	ADD_TO_LIST(c);
       } else
 	Add_Request(I_LOG, "*", F_WARN, "ircd:invalid NOTICE target %s via %s",
@@ -952,7 +1004,7 @@ static int ircd_notice_sb(INTERFACE *srv, struct peer_t *peer, unsigned short to
 
 BINDING_TYPE_ircd_server_cmd(ircd_squery_sb);
 static int ircd_squery_sb(INTERFACE *srv, struct peer_t *peer, unsigned short token,
-			  const char *sender, const char *lcsender, char *cmd,
+			  const char *sender, const char *lcsender,
 			  int argc, const char **argv)
 { /* args: <servicename> <text to be sent> */
   CLIENT *tcl;
@@ -972,7 +1024,7 @@ static int ircd_squery_sb(INTERFACE *srv, struct peer_t *peer, unsigned short to
   }
 #ifdef USE_SERVICES
   if (!CLIENT_IS_REMOTE(tcl))
-    New_Request(tcl->via->p.iface, 0, ":%s SQUERY %s :%s", sender, c, argv[1]);
+    New_Request(tcl->via->p.iface, 0, ":%s SQUERY %s :%s", sender, argv[0], argv[1]);
   else
 #endif
     _ircd_broadcast_msglist_old((IRCD *)srv->data, pp, token, sender, 1,
@@ -983,7 +1035,7 @@ static int ircd_squery_sb(INTERFACE *srv, struct peer_t *peer, unsigned short to
 #if IRCD_MULTICONNECT
 BINDING_TYPE_ircd_server_cmd(ircd_iprivmsg);
 static int ircd_iprivmsg(INTERFACE *srv, struct peer_t *peer, unsigned short token,
-			 const char *sender, const char *lcsender, char *cmd,
+			 const char *sender, const char *lcsender,
 			 int argc, const char **argv)
 { /* args: <id> <msgtarget> <text to be sent> */
   CLIENT *cl, *tcl;
@@ -991,7 +1043,8 @@ static int ircd_iprivmsg(INTERFACE *srv, struct peer_t *peer, unsigned short tok
   IRCD *ircd = (IRCD *)srv->data;
   char *c, *cnext;
   struct peer_priv *pp = peer->iface->data; /* it's really peer */
-  const char *tlist[MAXTARGETS];
+  unsigned int max_targets = _ircd_client_recvq[0];
+  const char *tlist[max_targets];
   size_t s = 0, s2 = 0;
   int id;
   char targets[MESSAGEMAX];
@@ -1011,7 +1064,7 @@ static int ircd_iprivmsg(INTERFACE *srv, struct peer_t *peer, unsigned short tok
   {
     if ((cnext = strchr(c, ',')))
       *cnext++ = 0;
-    if (s == MAXTARGETS) {
+    if (s == max_targets) {
       ERROR("ircd:too many targets of IPRIVMSG via %s, may lose message for %s",
 	    peer->dname, c);
       ircd_recover_done(pp, "Too many targets");
@@ -1023,7 +1076,7 @@ static int ircd_iprivmsg(INTERFACE *srv, struct peer_t *peer, unsigned short tok
 		    "ircd:not permitted channel message from %s via %s",
 		    sender, peer->dname);
       if (!(tch->mode & A_INVISIBLE)) {
-	_ircd_bmsgl_chan(tch->chan, cl, cl->user, cl->host, "PRIVMSG", argv[2]);
+	_ircd_bmsgl_chan(tch->chan, cl, cl->user, cl->vhost, "PRIVMSG", argv[2]);
 	ADD_TO_LIST(tch->chan->lcname);
       } else {
 	ERROR("ircd:IPRIVMSG via %s for channel %s ignored", peer->dname, c);
@@ -1035,7 +1088,7 @@ static int ircd_iprivmsg(INTERFACE *srv, struct peer_t *peer, unsigned short tok
 			 cl->cs->lcnick, "PRIVMSG", argv[2]);
       else
 	_ircd_bmsgl_mask(ircd, c, &c[1], 0, sender, cl->user,
-			 cl->host, "PRIVMSG", argv[2]);
+			 cl->vhost, "PRIVMSG", argv[2]);
       ADD_TO_LIST(c); //TODO: lowercase it?
     } else if ((tcl = _ircd_find_msg_target(c, pp))) {
       if (CLIENT_IS_SERVICE(tcl)) {
@@ -1047,7 +1100,7 @@ static int ircd_iprivmsg(INTERFACE *srv, struct peer_t *peer, unsigned short tok
 		      cl->cs->lcnick, c, argv[2]);
 	else
 	  New_Request(tcl->via->p.iface, 0, ":%s!%s@%s PRIVMSG %s :%s", sender,
-		      cl->user, cl->host, c, argv[2]);
+		      cl->user, cl->vhost, c, argv[2]);
 	if (tcl->umode & A_AWAY)
 	  ircd_do_unumeric(cl, RPL_AWAY, tcl, 0, tcl->away);
       } else
@@ -1067,7 +1120,7 @@ static int ircd_iprivmsg(INTERFACE *srv, struct peer_t *peer, unsigned short tok
 		      cl->cs->lcnick, c, argv[1]);
 	else
 	  Add_Request(I_PENDING, "*", 0, ":%s!%s@%s PRIVMSG %s :%s", sender,
-		      cl->user, cl->host, c, argv[1]);
+		      cl->user, cl->vhost, c, argv[1]);
 	ADD_TO_LIST(c);
       } else {
 	ERROR("ircd:invalid IPRIVMSG target %s via %s", c, peer->dname);
@@ -1086,7 +1139,7 @@ static int ircd_iprivmsg(INTERFACE *srv, struct peer_t *peer, unsigned short tok
 
 BINDING_TYPE_ircd_server_cmd(ircd_inotice);
 static int ircd_inotice(INTERFACE *srv, struct peer_t *peer, unsigned short token,
-			const char *sender, const char *lcsender, char *cmd,
+			const char *sender, const char *lcsender,
 			int argc, const char **argv)
 { /* args: <id> <msgtarget> <text to be sent> */
   CLIENT *cl, *tcl;
@@ -1094,7 +1147,8 @@ static int ircd_inotice(INTERFACE *srv, struct peer_t *peer, unsigned short toke
   IRCD *ircd = (IRCD *)srv->data;
   char *c, *cnext;
   struct peer_priv *pp = peer->iface->data; /* it's really peer */
-  const char *tlist[MAXTARGETS];
+  unsigned int max_targets = _ircd_client_recvq[0];
+  const char *tlist[max_targets];
   size_t s = 0, s2 = 0;
   int id;
   char targets[MESSAGEMAX];
@@ -1115,7 +1169,7 @@ static int ircd_inotice(INTERFACE *srv, struct peer_t *peer, unsigned short toke
   {
     if ((cnext = strchr(c, ',')))
       *cnext++ = 0;
-    if (s == MAXTARGETS)
+    if (s == max_targets)
       Add_Request(I_LOG, "*", F_WARN, "ircd:too many targets of INOTICE via "
 		  "%s, may lose message for %s", peer->dname, c);
     tch = ircd_find_member(ircd, c, NULL);
@@ -1125,7 +1179,7 @@ static int ircd_inotice(INTERFACE *srv, struct peer_t *peer, unsigned short toke
 		    "ircd:not permitted channel message from %s via %s",
 		    sender, peer->dname);
       if (!(tch->mode & A_INVISIBLE)) {
-	_ircd_bmsgl_chan(tch->chan, cl, cl->user, cl->host, "NOTICE", argv[2]);
+	_ircd_bmsgl_chan(tch->chan, cl, cl->user, cl->vhost, "NOTICE", argv[2]);
 	ADD_TO_LIST(tch->chan->lcname);
       } else
 	Add_Request(I_LOG, "*", F_WARN, "ircd:invalid INOTICE target %s via %s",
@@ -1136,7 +1190,7 @@ static int ircd_inotice(INTERFACE *srv, struct peer_t *peer, unsigned short toke
 			 cl->cs->lcnick, "NOTICE", argv[2]);
       else
 	_ircd_bmsgl_mask(ircd, c, &c[1], 0, sender, cl->user,
-			 cl->host, "NOTICE", argv[2]);
+			 cl->vhost, "NOTICE", argv[2]);
       ADD_TO_LIST(c); //TODO: lowercase it?
     } else if ((tcl = _ircd_find_msg_target(c, pp))) {
       if (CLIENT_IS_SERVICE(tcl)) {
@@ -1148,7 +1202,7 @@ static int ircd_inotice(INTERFACE *srv, struct peer_t *peer, unsigned short toke
 		      cl->cs->lcnick, c, argv[2]);
 	else
 	  New_Request(tcl->via->p.iface, 0, ":%s!%s@%s NOTICE %s :%s", sender,
-		      cl->user, cl->host, c, argv[2]);
+		      cl->user, cl->vhost, c, argv[2]);
       } else
 	ADD_TO_LIST(tcl->lcnick);
     } else {
@@ -1166,7 +1220,7 @@ static int ircd_inotice(INTERFACE *srv, struct peer_t *peer, unsigned short toke
 		      cl->cs->lcnick, c, argv[1]);
 	else
 	  Add_Request(I_PENDING, "*", 0, ":%s!%s@%s NOTICE %s :%s", sender,
-		      cl->user, cl->host, c, argv[1]);
+		      cl->user, cl->vhost, c, argv[1]);
 	ADD_TO_LIST(c);
       } else
 	Add_Request(I_LOG, "*", F_WARN, "ircd:invalid INOTICE target %s via %s",
@@ -1184,7 +1238,7 @@ static int ircd_inotice(INTERFACE *srv, struct peer_t *peer, unsigned short toke
 
 BINDING_TYPE_ircd_server_cmd(ircd_isquery);
 static int ircd_isquery(INTERFACE *srv, struct peer_t *peer, unsigned short token,
-			const char *sender, const char *lcsender, char *cmd,
+			const char *sender, const char *lcsender,
 			int argc, const char **argv)
 { /* args: <id> <servicename> <text to be sent> */
   CLIENT *tcl;
@@ -1210,7 +1264,7 @@ static int ircd_isquery(INTERFACE *srv, struct peer_t *peer, unsigned short toke
   }
 #ifdef USE_SERVICES
   if (!CLIENT_IS_REMOTE(tcl))
-    New_Request(tcl->via->p.iface, 0, ":%s SQUERY %s :%s", sender, c, argv[2]);
+    New_Request(tcl->via->p.iface, 0, ":%s SQUERY %s :%s", sender, argv[1], argv[2]);
   else
 #endif
   {
