@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2010-2012  Andrej N. Gritsenko <andrej@rep.kiev.ua>
+ * Copyright (C) 2010-2017  Andrej N. Gritsenko <andrej@rep.kiev.ua>
  *
  *     This program is free software; you can redistribute it and/or modify
  *     it under the terms of the GNU General Public License as published by
@@ -35,8 +35,11 @@ static char _ircd_motd_file[PATH_MAX+1] = "ircd.motd";
 static char _ircd_admin_info[SHORT_STRING] = "Not configured.";
 static char _ircd_admin_email[SHORT_STRING] = "lame@lame.org";
 extern char _ircd_description_string[]; /* in ircd.c */
+static long int _ircd_max_matches = 500;
+static long int _ircd_max_whois = 3;
 
 static struct bindtable_t *BTIrcdStatsReply;
+static struct bindtable_t *BTIrcdWhois;
 
 /* ---------------------------------------------------------------------------
  * Common internal functions.
@@ -140,7 +143,7 @@ int ircd_names_reply (CLIENT *me, CLIENT *cl, CHANNEL *ch, int done)
     {
       buf[p+s] = '\0';
       ircd_do_cnumeric (cl, RPL_NAMREPLY, ch, 0, buf);
-      if (done >= 0 && ++done == MAXMATCHES)
+      if (done >= 0 && ++done >= _ircd_max_matches)
       {
 	ircd_do_unumeric (cl, ERR_TOOMANYMATCHES, cl, 0, ch->name);
 	return done;
@@ -180,7 +183,7 @@ static int _ircd_names_reply_all (IRCD *ircd, CLIENT *me, CLIENT *cl, int i)
     ch = l->s.data;
     if (ch->users)
       i = ircd_names_reply (me, cl, ch, i);
-    if (i == MAXMATCHES)
+    if (i >= _ircd_max_matches)
       return ircd_do_unumeric (cl, RPL_ENDOFNAMES, cl, 0, "*");
   }
   /* list all visible users not on visible channels */
@@ -205,7 +208,7 @@ static int _ircd_names_reply_all (IRCD *ircd, CLIENT *me, CLIENT *cl, int i)
       if (s + strlen (tgt->nick) > x)
       {
 	ircd_do_unumeric (cl, RPL_NAMREPLY, cl, 0, buf);
-	if (i >= 0 && ++i == MAXMATCHES)
+	if (i >= 0 && ++i >= _ircd_max_matches)
 	{
 	  ircd_do_unumeric (cl, ERR_TOOMANYMATCHES, cl, 0, "*");
 	  return ircd_do_unumeric (cl, RPL_ENDOFNAMES, cl, 0, "*");
@@ -226,6 +229,7 @@ static inline int _ircd_query_names (IRCD *ircd, CLIENT *cl, struct peer_priv *v
   CLIENT *me;
   register MEMBER *cm;
   char *c, *cnext;
+  int done = 0;
 
   if (argc > 1)
   {
@@ -240,15 +244,15 @@ static inline int _ircd_query_names (IRCD *ircd, CLIENT *cl, struct peer_priv *v
   }
   me = ircd_find_client (NULL, NULL);
   if (argc == 0)			/* NAMES "*" */
-    return _ircd_names_reply_all (ircd, me, cl, -1);
+    return _ircd_names_reply_all (ircd, me, cl, done);
   c = (char *)argv[0];
   do {
     if ((cnext = strchr (c, ',')))
       *cnext++ = 0;
     if ((cm = ircd_find_member (ircd, c, NULL)) != NOSUCHCHANNEL)
-      ircd_names_reply (me, cl, cm->chan, -1);
+      done = ircd_names_reply (me, cl, cm->chan, done);
     ircd_do_unumeric (cl, RPL_ENDOFNAMES, cl, 0, c);
-  } while ((c = cnext));
+  } while ((c = cnext) && (done < _ircd_max_matches));
   return 1;
 }
 
@@ -259,8 +263,9 @@ static inline int _ircd_query_names (IRCD *ircd, CLIENT *cl, struct peer_priv *v
 	    peer->iface->data, argc, argv)
 
 BINDING_TYPE_ircd_client_cmd(ircd_names_cb);
-static int ircd_names_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick, char *user,
-			 char *host, int argc, const char **argv)
+static int ircd_names_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnick,
+			 const char *user, const char *host, const char *vhost,
+			 modeflag eum, int argc, const char **argv)
 {
   DO_CLIENT_QUERY (_ircd_query_names);
 }
@@ -288,7 +293,7 @@ static int ircd_names_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick, char
 
 BINDING_TYPE_ircd_server_cmd(ircd_names_sb);
 static int ircd_names_sb(INTERFACE *srv, struct peer_t *peer, unsigned short token,
-			 const char *sender, const char *lcsender, char *cmd,
+			 const char *sender, const char *lcsender,
 			 int argc, const char **argv)
 {
   DO_SERVER_QUERY (_ircd_query_names);
@@ -310,6 +315,7 @@ static inline int _ircd_query_list (IRCD *ircd, CLIENT *cl, struct peer_priv *vi
   CLIENT *me;
   register MEMBER *cm;
   char *c, *cnext;
+  int done = 0;
 
   if (argc > 1)
   {
@@ -322,14 +328,22 @@ static inline int _ircd_query_list (IRCD *ircd, CLIENT *cl, struct peer_priv *vi
 		 me->nick);
     return 1;
   }
-  me = ircd_find_client (NULL, NULL);
+//  me = ircd_find_client (NULL, NULL);
+  ircd_do_unumeric (cl, RPL_LISTSTART, cl, 0, NULL);
   if (argc == 0)			/* LIST "*" */
   {
     LEAF *l = NULL;
 
     while ((l = Next_Leaf (ircd->channels, l, NULL)))
       if (!(((CHANNEL *)l->s.data)->mode & A_PRIVATE)) /* hidden for listing */
+      {
 	_ircd_list_reply (cl, l->s.data);
+	if (++done >= _ircd_max_matches)
+	{
+	  ircd_do_unumeric (cl, ERR_TOOMANYMATCHES, cl, 0, "*");
+	  break;
+	}
+      }
     return ircd_do_unumeric (cl, RPL_LISTEND, cl, 0, NULL);
   }
   c = (char *)argv[0];
@@ -343,15 +357,16 @@ static inline int _ircd_query_list (IRCD *ircd, CLIENT *cl, struct peer_priv *vi
 }
 
 BINDING_TYPE_ircd_client_cmd(ircd_list_cb);
-static int ircd_list_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick, char *user,
-			char *host, int argc, const char **argv)
+static int ircd_list_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnick,
+			const char *user, const char *host, const char *vhost,
+			modeflag eum, int argc, const char **argv)
 {
   DO_CLIENT_QUERY (_ircd_query_list);
 }
 
 BINDING_TYPE_ircd_server_cmd(ircd_list_sb);
 static int ircd_list_sb(INTERFACE *srv, struct peer_t *peer, unsigned short token,
-			const char *sender, const char *lcsender, char *cmd,
+			const char *sender, const char *lcsender,
 			int argc, const char **argv)
 {
   DO_SERVER_QUERY (_ircd_query_list);
@@ -439,15 +454,16 @@ static inline int _ircd_query_motd (IRCD *ircd, CLIENT *cl, struct peer_priv *vi
 }
 
 BINDING_TYPE_ircd_client_cmd(ircd_motd_cb);
-static int ircd_motd_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick, char *user,
-			char *host, int argc, const char **argv)
+static int ircd_motd_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnick,
+			const char *user, const char *host, const char *vhost,
+			modeflag eum, int argc, const char **argv)
 {
   DO_CLIENT_QUERY (_ircd_query_motd);
 }
 
 BINDING_TYPE_ircd_server_cmd(ircd_motd_sb);
 static int ircd_motd_sb(INTERFACE *srv, struct peer_t *peer, unsigned short token,
-			const char *sender, const char *lcsender, char *cmd,
+			const char *sender, const char *lcsender,
 			int argc, const char **argv)
 {
   DO_SERVER_QUERY (_ircd_query_motd);
@@ -481,13 +497,34 @@ static inline int _ircd_query_lusers (IRCD *ircd, CLIENT *cl, struct peer_priv *
 #define COUNT_USERS(A,B,C,D) \
   for (l = A; l; l = l->prev) \
     COUNT_USERS_CHECK_STATE \
+    COUNT_USERS_CHECK_PATH \
     if (CLIENT_IS_SERVER(l->cl)) D++; \
     else if (CLIENT_IS_SERVICE(l->cl)) C++; \
     else if (l->cl->umode & (A_OP | A_HALFOP)) op++; \
     else B++
 #define COUNT_USERS_CHECK_STATE if (l->cl->via->p.state != P_TALK) continue; else
+#define COUNT_USERS_CHECK_PATH 
   COUNT_USERS (ircd->token[0]->c.lients, lu, ls, ll);
 #undef COUNT_USERS_CHECK_STATE
+#undef COUNT_USERS_CHECK_PATH
+  /* recheck counters */
+  if (ircd->token[0]->x.a.uc != (unsigned)(lu + op))
+  {
+    ERROR("ircd:/lusers: local users count mismatch on recount: %u != %d, fixing it",
+	  ircd->token[0]->x.a.uc, lu + op);
+    ircd->token[0]->x.a.uc = lu + op;
+  }
+  if (ls != 0)
+  {
+    ERROR("ircd:/lusers: found servers in clients on ME: %u != 0", ls);
+    ls = 0;
+  }
+  for (l = ircd->servers; l; l = l->prev)
+    if (CLIENT_IS_SERVER(l->cl))
+      ls++;
+    else
+      ERROR("ircd:/lusers: client %s in local servers list isn't a server",
+	    l->cl->nick);
   if (simple_match (smask, ircd->token[0]->lcnick) >= 0)
     gu = lu, gs = ls, gl = ll, lu += op, x = 1;
   else
@@ -496,12 +533,32 @@ static inline int _ircd_query_lusers (IRCD *ircd, CLIENT *cl, struct peer_priv *
   for (i = 1; i < ircd->s; i++)
     if (ircd->token[i] && simple_match (smask, ircd->token[i]->lcnick) >= 0)
     {
+      unsigned int su = 0, op_save = op;
+      if (!CLIENT_IS_SERVER(ircd->token[i]))
+      {
+	ERROR("ircd:token %d isn't a server but %s!!!", i, ircd->token[i]->nick);
+	ircd->token[i] = NULL;
+	continue;
+      }
 #define COUNT_USERS_CHECK_STATE 
-      COUNT_USERS (ircd->token[i]->c.lients, gu, gs, gl);
+#if IRCD_MULTICONNECT
+# define COUNT_USERS_CHECK_PATH if (l->cl->via != ircd->token[i]->via) continue; else
+#else
+# define COUNT_USERS_CHECK_PATH 
+#endif
+      COUNT_USERS (ircd->token[i]->c.lients, su, gs, gl);
 #undef COUNT_USERS_CHECK_STATE
+#undef COUNT_USERS_CHECK_PATH
       if (CLIENT_IS_LOCAL(ircd->token[i]))
 	ll++;
       x++;
+      gu += su;
+      if (ircd->token[i]->x.a.uc != (su + op - op_save))
+      {
+	ERROR("ircd:/lusers: users count on %s mismatch on recount: %u != %u, fixing it",
+	      ircd->token[i]->lcnick, ircd->token[i]->x.a.uc, su + op - op_save);
+	ircd->token[i]->x.a.uc = su + op - op_save;
+      }
     }
 #undef COUNT_USERS
   snprintf (buff, sizeof(buff), "There are %d users and %d services on %d servers",
@@ -519,89 +576,69 @@ static inline int _ircd_query_lusers (IRCD *ircd, CLIENT *cl, struct peer_priv *
   if (i > 0)
     ircd_do_unumeric (cl, RPL_LUSERCHANNELS, cl, (unsigned short)i, NULL);
   snprintf (buff, sizeof(buff), "I have %d clients and %d servers", lu, ll);
-  return ircd_do_unumeric (cl, RPL_LUSERME, cl, 0, buff);
+  ircd_do_unumeric (cl, RPL_LUSERME, cl, 0, buff);
+  if (argc > 0)
+    return 1;
+  snprintf (buff, sizeof(buff), "%u", ircd->lu);
+  ircd_do_unumeric (cl, RPL_LOCALUSERS, cl, (unsigned short)lu, buff);
+  snprintf (buff, sizeof(buff), "%u", ircd->gu);
+  return ircd_do_unumeric (cl, RPL_GLOBALUSERS, cl, (unsigned short)gu, buff);
 }
 
 BINDING_TYPE_ircd_client_cmd(ircd_lusers_cb);
-static int ircd_lusers_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick, char *user,
-			  char *host, int argc, const char **argv)
+static int ircd_lusers_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnick,
+			  const char *user, const char *host, const char *vhost,
+			  modeflag eum, int argc, const char **argv)
 {
   DO_CLIENT_QUERY (_ircd_query_lusers);
 }
 
 BINDING_TYPE_ircd_server_cmd(ircd_lusers_sb);
 static int ircd_lusers_sb(INTERFACE *srv, struct peer_t *peer, unsigned short token,
-			  const char *sender, const char *lcsender, char *cmd,
+			  const char *sender, const char *lcsender,
 			  int argc, const char **argv)
 {
   DO_SERVER_QUERY (_ircd_query_lusers);
 }
 
-const char ircd_version_flags[] = {
-#ifdef IRCD_TRACE_USERS
-'A',
-#endif
-#ifdef IRCD_SQUIT_YOUNGEST
-'B',
-#endif
+const char *ircd_version_flags =
 #ifdef IRCD_ENABLE_REHASH
-'E',
+"E"
 #endif
 #if IRCD_USES_ICONV
-'i',
-#endif
-#ifndef DEFAULT_INVISIBLE
-'I',
+"i"
 #endif
 #ifdef IRCD_ENABLE_DIE
-'J',
+"J"
 #endif
 #ifdef IRCD_ENABLE_KILL
-'K',
-#endif
-#ifdef WALLOP_ONLY_OPERS
-'l',
-#endif
-#ifdef IDLE_FROM_MSG
-'M',
-#endif
-#ifndef NO_SPARE_INVITES
-'N',
+"K"
 #endif
 #if IRCD_MULTICONNECT
-'o',
-#endif
-#ifdef IRCD_PUBLIC_TOPIC
-'O',
+"o"
 #endif
 #ifdef IRCD_ENABLE_RESTART
-'R',
+"R"
 #endif
 #ifdef USE_SERVICES
-'s',
+"s"
 #endif
 #ifdef IRCD_ENABLE_SUMMON
-'S',
+"S"
+#endif
+#ifdef TOPICWHOTIME
+"T"
 #endif
-//#ifdef TOPIC_WHO_TIME
-//'T',
-//#endif
 #ifdef IRCD_ENABLE_USERS
-'U',
-#endif
-#ifdef IRCD_STRICT_MODECMD
-'+',
-#endif
-#ifdef IRCD_IGNORE_MKEY_ARG
-'-',
+"U"
 #endif
 #ifdef ENABLE_IPV6
-'6',
+"6"
 #endif
 #if ! IRCD_STRICT_NAMES
-'8',
+"8"
 #endif
-'\0' };
+;
 
 static inline int _ircd_query_version (IRCD *ircd, CLIENT *cl, struct peer_priv *via,
 				       int argc, const char **argv)
@@ -621,15 +658,16 @@ static inline int _ircd_query_version (IRCD *ircd, CLIENT *cl, struct peer_priv
 }
 
 BINDING_TYPE_ircd_client_cmd(ircd_version_cb);
-static int ircd_version_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
-			   char *user, char *host, int argc, const char **argv)
+static int ircd_version_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnick,
+			   const char *user, const char *host, const char *vhost,
+			   modeflag eum, int argc, const char **argv)
 {
   DO_CLIENT_QUERY (_ircd_query_version);
 }
 
 BINDING_TYPE_ircd_server_cmd(ircd_version_sb);
 static int ircd_version_sb(INTERFACE *srv, struct peer_t *peer, unsigned short token,
-			   const char *sender, const char *lcsender, char *cmd,
+			   const char *sender, const char *lcsender,
 			   int argc, const char **argv)
 {
   DO_SERVER_QUERY (_ircd_query_version);
@@ -665,15 +703,16 @@ static inline int _ircd_query_stats (IRCD *ircd, CLIENT *cl, struct peer_priv *v
 }
 
 BINDING_TYPE_ircd_client_cmd(ircd_stats_cb);
-static int ircd_stats_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
-			 char *user, char *host, int argc, const char **argv)
+static int ircd_stats_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnick,
+			 const char *user, const char *host, const char *vhost,
+			 modeflag eum, int argc, const char **argv)
 {
   DO_CLIENT_QUERY (_ircd_query_stats);
 }
 
 BINDING_TYPE_ircd_server_cmd(ircd_stats_sb);
 static int ircd_stats_sb(INTERFACE *srv, struct peer_t *peer, unsigned short token,
-			 const char *sender, const char *lcsender, char *cmd,
+			 const char *sender, const char *lcsender,
 			 int argc, const char **argv)
 {
   DO_SERVER_QUERY (_ircd_query_stats);
@@ -725,15 +764,16 @@ static inline int _ircd_query_links (IRCD *ircd, CLIENT *cl, struct peer_priv *v
 }
 
 BINDING_TYPE_ircd_client_cmd(ircd_links_cb);
-static int ircd_links_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
-			 char *user, char *host, int argc, const char **argv)
+static int ircd_links_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnick,
+			 const char *user, const char *host, const char *vhost,
+			 modeflag eum, int argc, const char **argv)
 {
   DO_CLIENT_QUERY (_ircd_query_links);
 }
 
 BINDING_TYPE_ircd_server_cmd(ircd_links_sb);
 static int ircd_links_sb(INTERFACE *srv, struct peer_t *peer, unsigned short token,
-			 const char *sender, const char *lcsender, char *cmd,
+			 const char *sender, const char *lcsender,
 			 int argc, const char **argv)
 {
   DO_SERVER_QUERY (_ircd_query_links);
@@ -762,15 +802,16 @@ static inline int _ircd_query_time (IRCD *ircd, CLIENT *cl, struct peer_priv *vi
 }
 
 BINDING_TYPE_ircd_client_cmd(ircd_time_cb);
-static int ircd_time_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
-			char *user, char *host, int argc, const char **argv)
+static int ircd_time_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnick,
+			const char *user, const char *host, const char *vhost,
+			modeflag eum, int argc, const char **argv)
 {
   DO_CLIENT_QUERY (_ircd_query_time);
 }
 
 BINDING_TYPE_ircd_server_cmd(ircd_time_sb);
 static int ircd_time_sb(INTERFACE *srv, struct peer_t *peer, unsigned short token,
-			const char *sender, const char *lcsender, char *cmd,
+			const char *sender, const char *lcsender,
 			int argc, const char **argv)
 {
   DO_SERVER_QUERY (_ircd_query_time);
@@ -780,7 +821,7 @@ static inline int _ircd_query_connect (IRCD *ircd, CLIENT *cl, struct peer_priv
 				       int argc, const char **argv)
 { /* args: <target server> <port> [<remote server>] */
   if (argc < 2)
-    return ircd_do_unumeric (cl, ERR_NEEDMOREPARAMS, cl, 0, NULL);
+    return ircd_do_unumeric (cl, ERR_NEEDMOREPARAMS, cl, 0, "CONNECT");
   if (argc > 2)
   {
     register CLIENT *tgt = _ircd_find_by_mask (ircd, via, argv[2]);
@@ -801,15 +842,16 @@ static inline int _ircd_query_connect (IRCD *ircd, CLIENT *cl, struct peer_priv
 }
 
 BINDING_TYPE_ircd_client_cmd(ircd_connect_cb);
-static int ircd_connect_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
-			   char *user, char *host, int argc, const char **argv)
+static int ircd_connect_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnick,
+			   const char *user, const char *host, const char *vhost,
+			   modeflag eum, int argc, const char **argv)
 {
   DO_CLIENT_QUERY (_ircd_query_connect);
 }
 
 BINDING_TYPE_ircd_server_cmd(ircd_connect_sb);
 static int ircd_connect_sb(INTERFACE *srv, struct peer_t *peer, unsigned short token,
-			   const char *sender, const char *lcsender, char *cmd,
+			   const char *sender, const char *lcsender,
 			   int argc, const char **argv)
 {
   DO_SERVER_QUERY (_ircd_query_connect);
@@ -849,8 +891,11 @@ static inline int _ircd_query_trace (IRCD *ircd, CLIENT *cl, struct peer_priv *v
 		    CHARSET_UNICODE))
       flags[p++] = 'u';
 #endif
+    /* special support for Zlib connection */
+    if (Connchain_Check(&tgt->cs->via->p, 'Z') < 0)
+      flags[p++] = 'z';
     flags[p] = '\0';
-    snprintf (tstr, sizeof(tstr), "%s V%.4s%s %d %d %d", next->nick, next->away,
+    snprintf (tstr, sizeof(tstr), "%s V%c%s %d %d %d", next->nick, next->away[3],
 	      flags, (int)(Time - tgt->cs->via->started), via->p.iface->qsize,
 	      tgt->cs->via->p.iface->qsize);
     ircd_do_unumeric (cl, RPL_TRACELINK, tgt, O_DLEVEL, tstr);
@@ -862,15 +907,16 @@ static inline int _ircd_query_trace (IRCD *ircd, CLIENT *cl, struct peer_priv *v
 }
 
 BINDING_TYPE_ircd_client_cmd(ircd_trace_cb);
-static int ircd_trace_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
-			 char *user, char *host, int argc, const char **argv)
+static int ircd_trace_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnick,
+			 const char *user, const char *host, const char *vhost,
+			 modeflag eum, int argc, const char **argv)
 {
   DO_CLIENT_QUERY (_ircd_query_trace);
 }
 
 BINDING_TYPE_ircd_server_cmd(ircd_trace_sb);
 static int ircd_trace_sb(INTERFACE *srv, struct peer_t *peer, unsigned short token,
-			 const char *sender, const char *lcsender, char *cmd,
+			 const char *sender, const char *lcsender,
 			 int argc, const char **argv)
 {
   DO_SERVER_QUERY (_ircd_query_trace);
@@ -897,15 +943,16 @@ static inline int _ircd_query_admin (IRCD *ircd, CLIENT *cl, struct peer_priv *v
 }
 
 BINDING_TYPE_ircd_client_cmd(ircd_admin_cb);
-static int ircd_admin_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
-			 char *user, char *host, int argc, const char **argv)
+static int ircd_admin_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnick,
+			 const char *user, const char *host, const char *vhost,
+			 modeflag eum, int argc, const char **argv)
 {
   DO_CLIENT_QUERY (_ircd_query_admin);
 }
 
 BINDING_TYPE_ircd_server_cmd(ircd_admin_sb);
 static int ircd_admin_sb(INTERFACE *srv, struct peer_t *peer, unsigned short token,
-			 const char *sender, const char *lcsender, char *cmd,
+			 const char *sender, const char *lcsender,
 			 int argc, const char **argv)
 {
   DO_SERVER_QUERY (_ircd_query_admin);
@@ -933,9 +980,11 @@ static inline int _ircd_query_info (IRCD *ircd, CLIENT *cl, struct peer_priv *vi
   /* version, patchlevel, compilation time */
   ircd_do_unumeric (cl, RPL_INFO, cl, 0, "Module \"ircd\" of " PACKAGE " "
 		    VERSION ".");
+  ircd_do_unumeric (cl, RPL_INFO, cl, 0, "Designed in compliance with RFC2812"
+		    " using few extensions to improve usability.");
   ircd_do_unumeric (cl, RPL_INFO, cl, 0, "Birth date: " COMPILETIME ".");
   /* other info */
-  ircd_do_unumeric (cl, RPL_INFO, cl, 0, "Copyright 1999 - 2015 "
+  ircd_do_unumeric (cl, RPL_INFO, cl, 0, "Copyright 1999 - 2017 "
 		    "Andriy Grytsenko and others.");
   ircd_do_unumeric (cl, RPL_INFO, cl, 0, "This program is free software;"
 		    " you can redistribute it and/or modify it under the terms"
@@ -950,26 +999,28 @@ static inline int _ircd_query_info (IRCD *ircd, CLIENT *cl, struct peer_priv *vi
 }
 
 BINDING_TYPE_ircd_client_cmd(ircd_info_cb);
-static int ircd_info_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
-			char *user, char *host, int argc, const char **argv)
+static int ircd_info_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnick,
+			const char *user, const char *host, const char *vhost,
+			modeflag eum, int argc, const char **argv)
 {
   DO_CLIENT_QUERY (_ircd_query_info);
 }
 
 BINDING_TYPE_ircd_server_cmd(ircd_info_sb);
 static int ircd_info_sb(INTERFACE *srv, struct peer_t *peer, unsigned short token,
-			const char *sender, const char *lcsender, char *cmd,
+			const char *sender, const char *lcsender,
 			int argc, const char **argv)
 {
   DO_SERVER_QUERY (_ircd_query_info);
 }
 
-static void _ircd_do_whois (CLIENT *cl, CLIENT *tgt, CLIENT *me)
+static void _ircd_do_whois (IRCD *ircd, CLIENT *cl, CLIENT *tgt, CLIENT *me)
 {
   MEMBER *m;
   size_t ptr;
   char mc;
   char buf[IRCMSGLEN];
+  register struct binding_t *b;
 
   ircd_do_unumeric (cl, RPL_WHOISUSER, tgt, 0, tgt->fname);
   if (!CLIENT_IS_REMOTE (tgt)) {
@@ -1017,6 +1068,12 @@ static void _ircd_do_whois (CLIENT *cl, CLIENT *tgt, CLIENT *me)
     ircd_do_unumeric (cl, RPL_WHOISCHARSET, tgt, 0,
 		      Conversion_Charset (tgt->via->p.iface->conv));
 #endif
+  /* special support for SSL client connections */
+  if (tgt->umode & A_SSL)
+    ircd_do_unumeric (cl, RPL_WHOISSECURE, tgt, 0, NULL);
+  while ((b = Check_Bindtable (BTIrcdWhois, tgt->nick, U_ALL, U_ANYCH, NULL)))
+    if (b->name == NULL)
+      b->func (ircd, cl->nick, cl->umode, tgt->nick, tgt->host, tgt->vhost, tgt->umode);
 }
 
 static inline int _ircd_query_whois (IRCD *ircd, CLIENT *cl, struct peer_priv *via,
@@ -1049,10 +1106,10 @@ static inline int _ircd_query_whois (IRCD *ircd, CLIENT *cl, struct peer_priv *v
     tgt = ircd_find_client (c, via);
     if (tgt && !CLIENT_IS_SERVER(tgt))
     {
-      if (n++ >= MAXWHOIS)
+      if (n++ >= _ircd_max_whois)
 	ircd_do_unumeric (cl, ERR_TOOMANYTARGETS, tgt, 0, "Ignoring request.");
       else
-	_ircd_do_whois (cl, tgt, me);
+	_ircd_do_whois (ircd, cl, tgt, me);
     }
     else if (strpbrk (c, "*?"))
     {
@@ -1061,7 +1118,7 @@ static inline int _ircd_query_whois (IRCD *ircd, CLIENT *cl, struct peer_priv *v
 	NODE *t = ircd->clients;
 	LEAF *l = NULL;
 
-	while (n < MAXWHOIS && (l = Next_Leaf (t, l, NULL)))
+	while (n < _ircd_max_whois && (l = Next_Leaf (t, l, NULL)))
 	{
 	  tgt = l->s.data;
 	  if ((tgt->umode & (A_SERVER | A_SERVICE)) | tgt->hold_upto)
@@ -1070,7 +1127,7 @@ static inline int _ircd_query_whois (IRCD *ircd, CLIENT *cl, struct peer_priv *v
 	      _ircd_is_on_the_same_channel (cl, tgt))
 	    if (simple_match (c, tgt->lcnick) >= 0)
 	    {
-	      _ircd_do_whois (cl, tgt, me);
+	      _ircd_do_whois (ircd, cl, tgt, me);
 	      n++;
 	    }
 	}
@@ -1088,15 +1145,16 @@ static inline int _ircd_query_whois (IRCD *ircd, CLIENT *cl, struct peer_priv *v
 }
 
 BINDING_TYPE_ircd_client_cmd(ircd_whois_cb);
-static int ircd_whois_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
-			 char *user, char *host, int argc, const char **argv)
+static int ircd_whois_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnick,
+			 const char *user, const char *host, const char *vhost,
+			 modeflag eum, int argc, const char **argv)
 {
   DO_CLIENT_QUERY (_ircd_query_whois);
 }
 
 BINDING_TYPE_ircd_server_cmd(ircd_whois_sb);
 static int ircd_whois_sb(INTERFACE *srv, struct peer_t *peer, unsigned short token,
-			 const char *sender, const char *lcsender, char *cmd,
+			 const char *sender, const char *lcsender,
 			 int argc, const char **argv)
 {
   DO_SERVER_QUERY (_ircd_query_whois);
@@ -1143,8 +1201,8 @@ static inline int _ircd_query_whowas (IRCD *ircd, CLIENT *cl, struct peer_priv *
 		 argv[0], argv[1], tgt->nick);
     return 1;
   }
-  if (argc < 2 || (max = atoi (argv[1])) < 0 || max > 2 * MAXWHOIS)
-    max = 2 * MAXWHOIS;
+  if (argc < 2 || (max = atoi (argv[1])) < 0 || max > 2 * _ircd_max_whois)
+    max = 2 * _ircd_max_whois;
   unistrlower (targets, argv[0], sizeof(targets));
   for (c = targets; c; c = cnext)
   {
@@ -1159,16 +1217,17 @@ static inline int _ircd_query_whowas (IRCD *ircd, CLIENT *cl, struct peer_priv *
     {
       strfcpy (dummy.user, ww->user, sizeof(dummy.user));
       strfcpy (dummy.nick, ww->nick, sizeof(dummy.nick));
-      strfcpy (dummy.host, ww->host, sizeof(dummy.host));
+      strfcpy (dummy.vhost, ww->host, sizeof(dummy.vhost));
       dummy.via = NULL;
       dummy.umode = 0;
       ircd_do_unumeric (cl, RPL_WHOWASUSER, &dummy, 0, ww->fname);
       //TODO: whowas time!
       strfcpy (dummy.nick, ww->fromsrv, sizeof(dummy.nick));
       localtime_r(&ww->wason, &tmp);
-      strftime(dummy.host, sizeof(dummy.host), "%c", &tmp);
+      strftime(dummy.vhost, sizeof(dummy.vhost), "%c", &tmp);
       ircd_do_unumeric (cl, RPL_WHOISSERVER, &dummy, 0, ww->nick);
       n++;
+      DBG("ircd:whowas reply on %s: done %p, prev %p, %d of %d", dummy.nick, ww, ww->prev, n, max);
     }
     if (n)
       ircd_do_unumeric (cl, RPL_ENDOFWHOWAS, cl, 0, c);
@@ -1179,15 +1238,16 @@ static inline int _ircd_query_whowas (IRCD *ircd, CLIENT *cl, struct peer_priv *
 }
 
 BINDING_TYPE_ircd_client_cmd(ircd_whowas_cb);
-static int ircd_whowas_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
-			  char *user, char *host, int argc, const char **argv)
+static int ircd_whowas_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnick,
+			  const char *user, const char *host, const char *vhost,
+			  modeflag eum, int argc, const char **argv)
 {
   DO_CLIENT_QUERY (_ircd_query_whowas);
 }
 
 BINDING_TYPE_ircd_server_cmd(ircd_whowas_sb);
 static int ircd_whowas_sb(INTERFACE *srv, struct peer_t *peer, unsigned short token,
-			  const char *sender, const char *lcsender, char *cmd,
+			  const char *sender, const char *lcsender,
 			  int argc, const char **argv)
 {
   DO_SERVER_QUERY (_ircd_query_whowas);
@@ -1227,15 +1287,16 @@ static inline int _ircd_query_ping (IRCD *ircd, CLIENT *cl, struct peer_priv *vi
 }
 
 BINDING_TYPE_ircd_client_cmd(ircd_ping_cb);
-static int ircd_ping_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
-			char *user, char *host, int argc, const char **argv)
+static int ircd_ping_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnick,
+			const char *user, const char *host, const char *vhost,
+			modeflag eum, int argc, const char **argv)
 {
   DO_CLIENT_QUERY (_ircd_query_ping);
 }
 
 BINDING_TYPE_ircd_server_cmd(ircd_ping_sb);
 static int ircd_ping_sb(INTERFACE *srv, struct peer_t *peer, unsigned short token,
-			const char *sender, const char *lcsender, char *cmd,
+			const char *sender, const char *lcsender,
 			int argc, const char **argv)
 {
   DO_SERVER2_QUERY (_ircd_query_ping);
@@ -1262,15 +1323,16 @@ static inline int _ircd_query_pong (IRCD *ircd, CLIENT *cl, struct peer_priv *vi
 }
 
 BINDING_TYPE_ircd_client_cmd(ircd_pong_cb);
-static int ircd_pong_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
-			char *user, char *host, int argc, const char **argv)
+static int ircd_pong_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnick,
+			const char *user, const char *host, const char *vhost,
+			modeflag eum, int argc, const char **argv)
 {
   DO_CLIENT_QUERY (_ircd_query_pong);
 }
 
 BINDING_TYPE_ircd_server_cmd(ircd_pong_sb);
 static int ircd_pong_sb(INTERFACE *srv, struct peer_t *peer, unsigned short token,
-			const char *sender, const char *lcsender, char *cmd,
+			const char *sender, const char *lcsender,
 			int argc, const char **argv)
 {
   DO_SERVER2_QUERY (_ircd_query_pong);
@@ -1310,15 +1372,16 @@ static inline int _ircd_query_summon(IRCD *ircd, CLIENT *cl, struct peer_priv *v
 }
 
 BINDING_TYPE_ircd_client_cmd(ircd_summon_cb);
-static int ircd_summon_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
-			  char *user, char *host, int argc, const char **argv)
+static int ircd_summon_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnick,
+			  const char *user, const char *host, const char *vhost,
+			  modeflag eum, int argc, const char **argv)
 {
   DO_CLIENT_QUERY (_ircd_query_summon);
 }
 
 BINDING_TYPE_ircd_server_cmd(ircd_summon_sb);
 static int ircd_summon_sb(INTERFACE *srv, struct peer_t *peer, unsigned short token,
-			  const char *sender, const char *lcsender, char *cmd,
+			  const char *sender, const char *lcsender,
 			  int argc, const char **argv)
 {
   DO_SERVER_QUERY (_ircd_query_summon);
@@ -1361,15 +1424,16 @@ static inline int _ircd_query_users(IRCD *ircd, CLIENT *cl, struct peer_priv *vi
 }
 
 BINDING_TYPE_ircd_client_cmd(ircd_users_cb);
-static int ircd_users_cb(INTERFACE *srv, struct peer_t *peer, char *lcnick,
-			 char *user, char *host, int argc, const char **argv)
+static int ircd_users_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnick,
+			 const char *user, const char *host, const char *vhost,
+			 modeflag eum, int argc, const char **argv)
 {
   DO_CLIENT_QUERY (_ircd_query_users);
 }
 
 BINDING_TYPE_ircd_server_cmd(ircd_users_sb);
 static int ircd_users_sb(INTERFACE *srv, struct peer_t *peer, unsigned short token,
-			 const char *sender, const char *lcsender, char *cmd,
+			 const char *sender, const char *lcsender,
 			 int argc, const char **argv)
 {
   DO_SERVER_QUERY (_ircd_query_users);
@@ -1379,10 +1443,10 @@ static int ircd_users_sb(INTERFACE *srv, struct peer_t *peer, unsigned short tok
 /* ---------------------------------------------------------------------------
    "ircd-local-client" binding for user connection start */
 BINDING_TYPE_ircd_local_client(_igotc_lu_mo);
-static void _igotc_lu_mo (INTERFACE *srv, struct peer_t *peer)
+static void _igotc_lu_mo (INTERFACE *srv, struct peer_t *peer, modeflag um)
 {
-  ircd_lusers_cb (srv, peer, NULL, NULL, NULL, 0, NULL);
-  ircd_motd_cb (srv, peer, NULL, NULL, NULL, 0, NULL);
+  ircd_lusers_cb (srv, peer, NULL, NULL, NULL, NULL, A_ISON, 0, NULL);
+  ircd_motd_cb (srv, peer, NULL, NULL, NULL, NULL, A_ISON, 0, NULL);
 }
 
 
@@ -1422,6 +1486,8 @@ static void _icchg_ww(INTERFACE *srv, const char *from, const char *lcnick,
     IrcdWhowasUsed = IrcdWhowasPtr;
   }
   IrcdWhowasPtr++;
+  if (IrcdWhowasUsed < IrcdWhowasPtr)
+    IrcdWhowasUsed = IrcdWhowasPtr;
   wwp = Find_Key(IrcdWhowasTree, lcnick);
   if (wwp != NULL) {
     if (Delete_Key(IrcdWhowasTree, wwp->lcnick, wwp))
@@ -1432,6 +1498,7 @@ static void _icchg_ww(INTERFACE *srv, const char *from, const char *lcnick,
   }
   ww->next = NULL;
   ww->prev = wwp;
+  DBG("ircd:_ilostc_ww: adding %s: %p after %p, %d of %d", nick, ww, ww->prev, IrcdWhowasPtr, IrcdWhowasUsed);
   strfcpy(ww->nick, nick, sizeof(ww->nick));
   strfcpy(ww->lcnick, lcnick, sizeof(ww->lcnick));
   if (Insert_Key(&IrcdWhowasTree, ww->lcnick, ww, 1))
@@ -1452,7 +1519,7 @@ static CLIENT _istats_dummy_client = { .nick = "", .via = NULL };
 
 static void _istats_o_show_host (INTERFACE *tmp, char *host)
 {
-  strfcpy (_istats_dummy_client.host, host, sizeof(_istats_dummy_client.host));
+  strfcpy (_istats_dummy_client.vhost, host, sizeof(_istats_dummy_client.vhost));
   ircd_do_unumeric (_ircd_stats_client, RPL_STATSOLINE, &_istats_dummy_client,
 		    0, NULL);
 }
@@ -1461,7 +1528,7 @@ static void _istats_o_show_O (INTERFACE *tmp, char *oper)
 {
   lid_t lid = FindLID (oper);
 
-  strfcpy (_istats_dummy_client.user, oper, sizeof(_istats_dummy_client.user));
+  strfcpy (_istats_dummy_client.nick, oper, sizeof(_istats_dummy_client.nick));
   _ircd_list_receiver_show = &_istats_o_show_host;
   if (Get_Hostlist (tmp, lid))
     Get_Request();
@@ -1470,11 +1537,9 @@ static void _istats_o_show_O (INTERFACE *tmp, char *oper)
 BINDING_TYPE_ircd_stats_reply(_istats_o);
 static void _istats_o (INTERFACE *srv, const char *rq, modeflag umode)
 {
-  char n[NAMEMAX+2];
+  const char *n = ((IRCD *)srv->data)->sub->name;
   INTERFACE *tmp;
 
-  n[0] = '@';
-  strfcpy (&n[1], srv->name, sizeof(n)-1);
   tmp = Add_Iface (I_TEMP, NULL, NULL, &_ircd_qlist_r, NULL);
   _ircd_list_receiver_show = &_istats_o_show_O;
   Set_Iface (tmp);
@@ -1503,6 +1568,114 @@ static void _istats_u (INTERFACE *srv, const char *rq, modeflag umode)
 		    0, buf);
 }
 
+/* RFC1459 mentions also stats c, h, i, k, y but those are optional */
+static void _istats_c_show_host (INTERFACE *tmp, char *host)
+{
+  char *c = strchr (host, ':');
+  unsigned short port = 0;
+
+  if (c == NULL)
+    c = strchr (host, '@');
+  if (c)
+  {
+    size_t hs = c - host;
+    if (hs >= sizeof(_istats_dummy_client.vhost))
+      hs = sizeof(_istats_dummy_client.vhost) - 1;
+    strfcpy (_istats_dummy_client.vhost, host, hs + 1);
+    if (*c == ':')
+      c = strchr (c, '@');
+    if (!_istats_dummy_client.vhost[0])
+      strcpy (_istats_dummy_client.vhost, "*");
+  }
+  else
+  {
+    c = host;
+    strcpy (_istats_dummy_client.vhost, "*@");
+  }
+  strfcat (_istats_dummy_client.vhost, c, sizeof(_istats_dummy_client.vhost));
+  /* _istats_dummy_client.vhost contains now user@host/port */
+  c = strchr (_istats_dummy_client.vhost, '/');
+  if (c)
+  {
+    *c++ = '\0';
+    port = (unsigned short)strtoul(c, NULL, 10);
+  }
+  ircd_do_unumeric (_ircd_stats_client, RPL_STATSCLINE, &_istats_dummy_client,
+		    port, NULL);
+}
+
+static void _istats_c_show_S (INTERFACE *tmp, char *serv)
+{
+  lid_t lid = FindLID (serv);
+
+  strfcpy (_istats_dummy_client.nick, serv, sizeof(_istats_dummy_client.nick));
+  _ircd_list_receiver_show = &_istats_c_show_host;
+  if (Get_Hostlist (tmp, lid))
+    Get_Request();
+}
+
+BINDING_TYPE_ircd_stats_reply(_istats_c);
+static void _istats_c (INTERFACE *srv, const char *rq, modeflag umode)
+{
+  const char *n = ((IRCD *)srv->data)->sub->name;
+  INTERFACE *tmp;
+
+  if (!(umode & (A_OP | A_HALFOP)))
+  {
+    ircd_do_unumeric (_ircd_stats_client, ERR_NOPRIVILEGES, _ircd_stats_client, 0, NULL);
+    return;
+  }
+  tmp = Add_Iface (I_TEMP, NULL, NULL, &_ircd_qlist_r, NULL);
+  _ircd_list_receiver_show = &_istats_c_show_S;
+  Set_Iface (tmp);
+  if (Get_Clientlist (tmp, U_UNSHARED, n, "*"))
+    Get_Request(); /* it will do recurse itself */
+  Unset_Iface();
+  tmp->ift = I_DIED;
+}
+
+static void _istats_h_show (INTERFACE *tmp, char *serv)
+{
+  struct clrec_t *clr = Lock_Clientrecord (serv);
+  char *hub, *c1, *c2;
+
+  if (clr)
+  {
+    strfcpy (_istats_dummy_client.nick, serv, sizeof(_istats_dummy_client.nick));
+    hub = safe_strdup (Get_Field (clr, "hub", NULL));
+    Unlock_Clientrecord (clr);
+    if (hub)
+    {
+      for (c1 = hub; *c1; c1 = c2)
+      {
+	c2 = gettoken (c1, NULL);
+	ircd_do_unumeric (_ircd_stats_client, RPL_STATSHLINE, &_istats_dummy_client, 0, c1);
+      }
+      free (hub);
+    }
+  }
+}
+
+BINDING_TYPE_ircd_stats_reply(_istats_h);
+static void _istats_h (INTERFACE *srv, const char *rq, modeflag umode)
+{
+  const char *n = ((IRCD *)srv->data)->sub->name;
+  INTERFACE *tmp;
+
+  if (!(umode & (A_OP | A_HALFOP)))
+  {
+    ircd_do_unumeric (_ircd_stats_client, ERR_NOPRIVILEGES, _ircd_stats_client, 0, NULL);
+    return;
+  }
+  tmp = Add_Iface (I_TEMP, NULL, NULL, &_ircd_qlist_r, NULL);
+  _ircd_list_receiver_show = &_istats_h_show;
+  Set_Iface (tmp);
+  if (Get_Clientlist (tmp, U_UNSHARED, n, "*"))
+    Get_Request(); /* it will do recurse itself */
+  Unset_Iface();
+  tmp->ift = I_DIED;
+}
+
 
 /* ---------------------------------------------------------------------------
  * Common external functions.
@@ -1514,6 +1687,8 @@ void ircd_queries_proto_end (void)
   UnregisterVariable ("ircd-motd-file");
   UnregisterVariable ("ircd-admin-info");
   UnregisterVariable ("ircd-admin-email");
+  UnregisterVariable ("ircd-max-matches");
+  UnregisterVariable ("ircd-max-whois");
   FREE (&IrcdMotd);
   IrcdMotdSize = 0;
   Delete_Binding ("ircd-client-cmd", &ircd_names_cb, NULL);
@@ -1556,6 +1731,8 @@ void ircd_queries_proto_end (void)
   Delete_Binding ("ircd-client", (Function)&_icchg_ww, NULL);
   Delete_Binding ("ircd-stats-reply", (Function)&_istats_o, NULL);
   Delete_Binding ("ircd-stats-reply", (Function)&_istats_u, NULL);
+  Delete_Binding ("ircd-stats-reply", (Function)&_istats_c, NULL);
+  Delete_Binding ("ircd-stats-reply", (Function)&_istats_h, NULL);
   Destroy_Tree (&IrcdWhowasTree, NULL);
   FREE (&IrcdWhowasArray);
 }
@@ -1567,6 +1744,8 @@ void ircd_queries_register (void)
 		  sizeof(_ircd_admin_info), 0);
   RegisterString ("ircd-admin-email", _ircd_admin_email,
 		  sizeof(_ircd_admin_email), 0);
+  RegisterInteger ("ircd-max-matches", &_ircd_max_matches);
+  RegisterInteger ("ircd-max-whois", &_ircd_max_whois);
 }
 
 void ircd_queries_proto_start (void)
@@ -1574,6 +1753,7 @@ void ircd_queries_proto_start (void)
   ircd_queries_register();
   //TODO: fail to start if no admin email set?
   BTIrcdStatsReply = Add_Bindtable ("ircd-stats-reply", B_KEYWORD);
+  BTIrcdWhois = Add_Bindtable ("ircd-whois", B_MASK);
   Add_Binding ("ircd-client-cmd", "names", 0, 0, &ircd_names_cb, NULL);
   Add_Binding ("ircd-client-cmd", "list", 0, 0, &ircd_list_cb, NULL);
   Add_Binding ("ircd-client-cmd", "motd", 0, 0, &ircd_motd_cb, NULL);
@@ -1614,6 +1794,8 @@ void ircd_queries_proto_start (void)
   Add_Binding ("ircd-client", "*", 0, 0, (Function)&_icchg_ww, NULL);
   Add_Binding ("ircd-stats-reply", "o", 0, 0, (Function)&_istats_o, NULL);
   Add_Binding ("ircd-stats-reply", "u", 0, 0, (Function)&_istats_u, NULL);
+  Add_Binding ("ircd-stats-reply", "c", 0, 0, (Function)&_istats_c, NULL);
+  Add_Binding ("ircd-stats-reply", "h", 0, 0, (Function)&_istats_h, NULL);
   _ircd_time_started = Time;
 //TODO: add bindtable to add chars into ircd_version_flags?
 }
