@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2010-2011  Andrej N. Gritsenko <andrej@rep.kiev.ua>
+ * Copyright (C) 2010-2016  Andrej N. Gritsenko <andrej@rep.kiev.ua>
  *
  *     This program is free software; you can redistribute it and/or modify
  *     it under the terms of the GNU General Public License as published by
@@ -68,7 +68,7 @@
 #endif
 
 #if IRCD_MULTICONNECT /* don't send back to sender by cycle */
-#define __CHECK_TRANSIT__(_a_) if (L->cl->x.token != _a_)
+#define __CHECK_TRANSIT__(_a_) if (L->cl->x.a.token != _a_)
 #else /* we never can send to sender if not back on RFC2813 server */
 #define __CHECK_TRANSIT__(_a_)
 #endif
@@ -98,6 +98,14 @@
       __TRANSIT__ L->cl->via->p.iface->ift |= I_PENDING; \
   Add_Request (I_PENDING, "*", 0, __VA_ARGS__); } while(0)
 #if IRCD_MULTICONNECT
+/* sends to every server; args: ircd, from_peer, but, message...
+   similar to ircd_sendto_servers_all, but not send to server $but */
+#define ircd_sendto_servers_all_but(i,a,b,...) do {\
+  register LINK *L; \
+  for (L = (i)->servers; L; L = L->prev) \
+    if (L->cl->via != a && L->cl != b) \
+      __TRANSIT__ L->cl->via->p.iface->ift |= I_PENDING; \
+  Add_Request (I_PENDING, "*", 0, __VA_ARGS__); } while(0)
 /* sends to every new type server */
 #define ircd_sendto_servers_new(a,b,...) do {\
   register LINK *L; \
@@ -147,11 +155,12 @@
       L->cl->via->p.iface->ift |= I_PENDING; \
       ircd_add_ack (L->cl->via, a, b); } \
   Add_Request (I_PENDING, "*", 0, __VA_ARGS__); } while(0)
-/* sends to every server with ack; args the same */
+/* sends to every server with ack; args the same;
+   also test to not send to $who as required for SQUIT */
 #define ircd_sendto_servers_all_ack(i,a,b,c,...) do {\
   LINK *L; \
   for (L = (i)->servers; L; L = L->prev) \
-    __TRANSIT__ if (L->cl->via != c) { \
+    __TRANSIT__ if (L->cl->via != c && L->cl != a) { \
       L->cl->via->p.iface->ift |= I_PENDING; \
       if (L->cl->umode & A_MULTI) \
 	ircd_add_ack (L->cl->via, a, b); } \
@@ -166,6 +175,8 @@
 	ircd_add_ack (L->cl->via, a, b); } \
   Add_Request (I_PENDING, "*", 0, __VA_ARGS__); } while(0)
 #else
+#define ircd_sendto_servers_all_but(i,a,b,...) \
+  ircd_sendto_servers_all(i,a,__VA_ARGS__)
 #define ircd_sendto_servers_new(a,...)
 #define ircd_sendto_servers_ack(a,...)
 #define ircd_sendto_servers_mask_new(a,...)
@@ -179,7 +190,7 @@
 #endif
 
 //TODO: implement IWALLOPS too
-/* broadcasts WALLOPS; args: ircd, from peer, sender/me, text, ... */
+/* broadcasts WALLOPS; args: ircd, from peer, sender/me, message... */
 #define ircd_sendto_wallops(i,a,b,c,...) do { \
   __attribute__((unused)) register const char *me = ircd_mark_wallops(); \
   register LINK *L; \
@@ -189,5 +200,44 @@
   Add_Request (I_PENDING|I_LOG, "*", F_WALL, ":%s WALLOPS :" c, \
 	       b, __VA_ARGS__); } while(0)
 
+#ifdef USE_SERVICES
+#ifdef __IN_IRCD_C /* in ircd.c */
+# define SERVICES_LIST_PATH(a) ME.c.lients
+#else
+# define SERVICES_LIST_PATH(a) (a)->token[0]->c.lients
+#endif
+/* mark local services to send; args: ircd, flags */
+#define ircd_sendto_services_mark_prefix(a,b) do { \
+  register LINK *L; \
+  for (L = SERVICES_LIST_PATH(a); L; L = L->prev) \
+    if (CLIENT_IS_SERVICE(L->cl) && \
+	(SERVICE_FLAGS(L->cl) & (b)) && \
+	(SERVICE_FLAGS(L->cl) & SERVICE_WANT_PREFIX)) \
+      L->cl->via->p.iface->ift |= I_PENDING; } while(0)
+#define ircd_sendto_services_mark_nick(a,b) do { \
+  register LINK *L; \
+  for (L = SERVICES_LIST_PATH(a); L; L = L->prev) \
+    if (CLIENT_IS_SERVICE(L->cl) && \
+	(SERVICE_FLAGS(L->cl) & (b)) && \
+	!(SERVICE_FLAGS(L->cl) & SERVICE_WANT_PREFIX)) \
+      L->cl->via->p.iface->ift |= I_PENDING; } while(0)
+#define ircd_sendto_services_mark_all(a,b) do { \
+  register LINK *L; \
+  for (L = SERVICES_LIST_PATH(a); L; L = L->prev) \
+    if (CLIENT_IS_SERVICE(L->cl) && \
+	(SERVICE_FLAGS(L->cl) & (b))) \
+      L->cl->via->p.iface->ift |= I_PENDING; } while(0)
+/* send message to local services; args: ircd, flags, message... */
+#define ircd_sendto_services_prefix(a,b,...) do { \
+  ircd_sendto_services_mark_prefix(a,b); \
+  Add_Request (I_PENDING, "*", 0, __VA_ARGS__); } while(0)
+#define ircd_sendto_services_nick(a,b,...) do { \
+  ircd_sendto_services_mark_nick(a,b); \
+  Add_Request (I_PENDING, "*", 0, __VA_ARGS__); } while(0)
+#define ircd_sendto_services_all(a,b,...) do { \
+  ircd_sendto_services_mark_all(a,b); \
+  Add_Request (I_PENDING, "*", 0, __VA_ARGS__); } while(0)
+#endif /* USE_SERVICES */
+
 #define __TRANSIT__ /* no transit by default */
 #endif
