@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2011-2014  Andrej N. Gritsenko <andrej@rep.kiev.ua>
+ * Copyright (C) 2011-2017  Andrej N. Gritsenko <andrej@rep.kiev.ua>
  *
  *     This program is free software; you can redistribute it and/or modify
  *     it under the terms of the GNU General Public License as published by
@@ -90,14 +90,44 @@ void ircd_drop_ack (IRCD *ircd, struct peer_priv *link)
 /* finds ack for client, on channel or global */
 ACK *ircd_check_ack (struct peer_priv *link, CLIENT *who, CHANNEL *where)
 {
-  register ACK *ack;
+  register ACK *ack, *possible = NULL;
 
   for (ack = link->acks; ack; ack = ack->next)
-    if (!ack->contrary && ack->who == who && ack->where == where)
+    if (ack->contrary) {
+      if (ack->who == who && ack->where == where)
+        possible = ack;
+    } else if (ack->who == who && ack->where == where)
       return ack;
+    else
+      return possible;
   return NULL;
 }
 
+static ACK *ircd_find_ack(struct peer_priv *link, const char *who, const char *where)
+{
+  register ACK *ack, *possible = NULL;
+
+  for (ack = link->acks; ack; ack = ack->next)
+  {
+    if (where == NULL && ack->where == NULL && strcmp(who, ack->who->nick) == 0) {
+      if (ack->contrary)
+	possible = ack;
+      else
+	return ack;
+    } else if (where != NULL && ack->where != NULL &&
+	       strcmp(who, ack->who->nick) == 0 &&
+	       ((ack->where == CHANNEL0 && strcmp(where, "0") == 0) ||
+		(ack->where != CHANNEL0 && strcmp(where, ack->where->name) == 0))) {
+      if (ack->contrary)
+	possible = ack;
+      else
+	return ack;
+    } else if (!ack->contrary)
+      break;
+  }
+  return possible;
+}
+
 
 /* -- ids management ------------------------------------------------------ */
 static int localid = 0;
@@ -254,7 +284,7 @@ int ircd_test_id(CLIENT *cl, int id)
 
 BINDING_TYPE_ircd_server_cmd(ircd_quit_sb);
 static int ircd_quit_sb(INTERFACE *srv, struct peer_t *peer, unsigned short token,
-			const char *sender, const char *lcsender, char *cmd,
+			const char *sender, const char *lcsender,
 			int argc, const char **argv)
 { /* args: [<Quit Message>] */
   CLIENT *cl;
@@ -265,18 +295,23 @@ static int ircd_quit_sb(INTERFACE *srv, struct peer_t *peer, unsigned short toke
   else
     msg = argv[0];
   cl = _ircd_find_client_lc((IRCD *)srv->data, lcsender);
-#ifdef USE_SERVICES
-  //TODO: notify services
-#endif
 #if IRCD_MULTICONNECT
   if (((struct peer_priv *)peer->iface->data)->link->cl->umode & A_MULTI)
     New_Request(peer->iface, 0, "ACK QUIT %s", sender);
+#endif
+#ifdef USE_SERVICES
+  ircd_sendto_services_mark_nick((IRCD *)srv->data,
+				 SERVICE_WANT_QUIT | SERVICE_WANT_RQUIT);
 #endif
   ircd_sendto_servers_all_ack((IRCD *)srv->data, cl, NULL, peer->iface->data,
 			      ":%s QUIT :%s", cl->nick, msg);
   ircd_prepare_quit(cl, cl->via, msg);
+#ifdef USE_SERVICES
+  ircd_sendto_services_mark_prefix((IRCD *)srv->data,
+				   SERVICE_WANT_QUIT | SERVICE_WANT_RQUIT);
+#endif
   Add_Request(I_PENDING, "*", 0, ":%s!%s@%s QUIT :%s", cl->nick, cl->user,
-	      cl->host, msg);
+	      cl->vhost, msg);
   cl->hold_upto = Time;
   cl->host[0] = 0;			/* for collision check */
   return 1;
@@ -284,7 +319,7 @@ static int ircd_quit_sb(INTERFACE *srv, struct peer_t *peer, unsigned short toke
 
 BINDING_TYPE_ircd_server_cmd(ircd_squit_sb);
 static int ircd_squit_sb(INTERFACE *srv, struct peer_t *peer, unsigned short token,
-			 const char *sender, const char *lcsender, char *cmd,
+			 const char *sender, const char *lcsender,
 			 int argc, const char **argv)
 { /* args: <server> <comment> */
   CLIENT *cl, *tgt;
@@ -318,15 +353,18 @@ static int ircd_squit_sb(INTERFACE *srv, struct peer_t *peer, unsigned short tok
     if (pp->link->cl->umode & A_MULTI)
       New_Request(peer->iface, 0, "ACK SQUIT %s", argv[0]);
 #endif
-    ack = ircd_check_ack(pp, tgt, NULL);
-    if (ack != NULL) {
-      ack->contrary = 1;
-      return (1); /* ignore the message */
-    }
     for (l = cl->c.lients; l; l = l->prev)
       if (l->cl == tgt)
 	break;
     if (l == NULL) {			/* ambiguous sender */
+#if IRCD_MULTICONNECT
+      /* or it might be a duplicate still */
+      ack = ircd_check_ack(pp, tgt, NULL);
+      if (ack != NULL) {
+	ack->contrary = 1;
+	return (1); /* ignore the message */
+      }
+#endif
       ircd_do_squit(pp->link, pp, "Invalid SQUIT message");
       return 0;				/* kill our peer instead */
     }
@@ -375,10 +413,7 @@ static inline int _ircd_join_0_remote(IRCD *ircd, struct peer_priv *bysrv,
 			     ch->name);
     else
       ircd_sendto_chan_local(ch, ":%s!%s@%s PART %s :%s", cl->nick,
-			     cl->user, cl->host, ch->name, key);
-#ifdef USE_SERVICES
-    //TODO: inform services
-#endif
+			     cl->user, cl->vhost, ch->name, key);
     ircd_del_from_channel(ircd, cl->c.hannels, 0);
   }
   ircd_sendto_servers_all_ack(ircd, cl, CHANNEL0, bysrv, ":%s JOIN 0 :%s",
@@ -388,7 +423,7 @@ static inline int _ircd_join_0_remote(IRCD *ircd, struct peer_priv *bysrv,
 
 BINDING_TYPE_ircd_server_cmd(ircd_join_sb);
 static int ircd_join_sb(INTERFACE *srv, struct peer_t *peer, unsigned short token,
-			const char *sender, const char *lcsender, char *cmd,
+			const char *sender, const char *lcsender,
 			int argc, const char **argv)
 { /* args: <channel>[^G<modes>][,<channel>[^G<modes>]]
            0 [reason] */
@@ -463,6 +498,10 @@ static int ircd_join_sb(INTERFACE *srv, struct peer_t *peer, unsigned short toke
       else
 	m += ptr;
     } else
+#if !IRCD_MULTICONNECT
+    /* for multiconnected server don't set error */
+    if (!(pp->link->cl->umode & A_MULTI))
+#endif
       err = 1;
     if (*c == ',')
       c++;
@@ -481,7 +520,7 @@ static int ircd_join_sb(INTERFACE *srv, struct peer_t *peer, unsigned short toke
 
 BINDING_TYPE_ircd_server_cmd(ircd_njoin);
 static int ircd_njoin(INTERFACE *srv, struct peer_t *peer, unsigned short token,
-		      const char *sender, const char *lcsender, char *cmd,
+		      const char *sender, const char *lcsender,
 		      int argc, const char **argv)
 { /* args: <channel> [@@|@|+]<nickname>[,[@@|@|+]<nickname>...] */
   MEMBER *tgt;
@@ -504,7 +543,7 @@ static int ircd_njoin(INTERFACE *srv, struct peer_t *peer, unsigned short token,
   if (ircd_char2mode(srv, peer->dname, sender, argv[0], '\0') & A_INVISIBLE ||
       ((c = strchr(argv[0], ':')) != NULL &&
        (cl = ircd_find_client(NULL, NULL)) &&
-       simple_match(&c[1], cl->lcnick)) < 0) {
+       simple_match(&c[1], cl->lcnick) < 0)) {
     ERROR("ircd:invalid NJOIN channel %s via %s", argv[0], peer->dname);
     return ircd_recover_done(pp, "Invalid NJOIN channel");
   }
@@ -580,7 +619,7 @@ static int ircd_njoin(INTERFACE *srv, struct peer_t *peer, unsigned short token,
 
 BINDING_TYPE_ircd_server_cmd(ircd_part_sb);
 static int ircd_part_sb(INTERFACE *srv, struct peer_t *peer, unsigned short token,
-			const char *sender, const char *lcsender, char *cmd,
+			const char *sender, const char *lcsender,
 			int argc, const char **argv)
 { /* args: <channel>[,<channel> ...] [<Part Message>] */
   CLIENT *cl;
@@ -619,6 +658,14 @@ static int ircd_part_sb(INTERFACE *srv, struct peer_t *peer, unsigned short toke
     *t = '\0';
     memb = ircd_find_member ((IRCD *)srv->data, chname, NULL);
     if (memb == NOSUCHCHANNEL) {
+#if !IRCD_MULTICONNECT
+      /* for multiconnected server don't set error */
+      if (pp->link->cl->umode & A_MULTI) {
+	DBG("ircd:got PART from %s for %s on nonexistent channel %s",
+	    peer->dname, sender, chname);
+	continue;
+      }
+#endif
       ERROR("ircd:got PART from %s for %s on nonexistent channel %s",
 	    peer->dname, sender, chname);
       ircd_recover_done(pp, "PART for nonexistent channel");
@@ -646,10 +693,7 @@ static int ircd_part_sb(INTERFACE *srv, struct peer_t *peer, unsigned short toke
 			     memb->chan->name);
     else
       ircd_sendto_chan_local(memb->chan, ":%s!%s@%s PART %s :%s", cl->nick,
-			     cl->user, cl->host, memb->chan->name, msg);
-#ifdef USE_SERVICES
-    //TODO: notify services
-#endif
+			     cl->user, cl->vhost, memb->chan->name, msg);
     if ((cmask = strchr (chname, ':'))) /* notify servers */ {
       cmask++; /* not put '++' into macro below */
       ircd_sendto_servers_mask_all_ack((IRCD *)srv->data, cl, memb->chan, pp,
@@ -672,7 +716,7 @@ _ircd_do_stopic(IRCD *ircd, const char *via, const char *sender,
 
 BINDING_TYPE_ircd_server_cmd(ircd_topic_sb);
 static int ircd_topic_sb(INTERFACE *srv, struct peer_t *peer, unsigned short token,
-			 const char *sender, const char *lcsender, char *cmd,
+			 const char *sender, const char *lcsender,
 			 int argc, const char **argv)
 { /* args: <channel> <topic> */
   CLIENT *cl;
@@ -698,7 +742,7 @@ static int ircd_topic_sb(INTERFACE *srv, struct peer_t *peer, unsigned short tok
 
 BINDING_TYPE_ircd_server_cmd(ircd_invite_sb);
 static int ircd_invite_sb(INTERFACE *srv, struct peer_t *peer, unsigned short token,
-			  const char *sender, const char *lcsender, char *cmd,
+			  const char *sender, const char *lcsender,
 			  int argc, const char **argv)
 { /* args: <nickname> <channel> */
   CLIENT *cl, *tgt;
@@ -717,7 +761,6 @@ static int ircd_invite_sb(INTERFACE *srv, struct peer_t *peer, unsigned short to
   tgt = ircd_find_client (argv[0], pp);
   if (!tgt || (tgt->umode & (A_SERVER|A_SERVICE)))
     return ircd_do_unumeric (cl, ERR_NOSUCHNICK, cl, 0, argv[0]);
-    //TODO: what if cl is a server?
   me = ircd_find_member ((IRCD *)srv->data, argv[1], NULL);
   if (me != NOSUCHCHANNEL)
   {
@@ -738,7 +781,7 @@ static int ircd_invite_sb(INTERFACE *srv, struct peer_t *peer, unsigned short to
   if (CLIENT_IS_REMOTE(tgt))
     ircd_sendto_one (tgt, ":%s INVITE %s %s", sender, argv[0], argv[1]);
   else
-    ircd_sendto_one (tgt, ":%s!%s@%s INVITE %s %s", sender, cl->user, cl->host,
+    ircd_sendto_one (tgt, ":%s!%s@%s INVITE %s %s", sender, cl->user, cl->vhost,
 		     argv[0], argv[1]);
   if (!(cl->umode & A_SERVICE) && tgt->away[0])
     ircd_do_unumeric (cl, RPL_AWAY, tgt, 0, tgt->away);
@@ -747,7 +790,7 @@ static int ircd_invite_sb(INTERFACE *srv, struct peer_t *peer, unsigned short to
 
 BINDING_TYPE_ircd_server_cmd(ircd_kick_sb);
 static int ircd_kick_sb(INTERFACE *srv, struct peer_t *peer, unsigned short token,
-			const char *sender, const char *lcsender, char *cmd,
+			const char *sender, const char *lcsender,
 			int argc, const char **argv)
 { /* args: <channel>[,<channel> ...] <user>[,<user> ...] [<comment>] */
   CLIENT *cl, *tgt;
@@ -843,10 +886,7 @@ static int ircd_kick_sb(INTERFACE *srv, struct peer_t *peer, unsigned short toke
 		      chn, lcl, reason);
       } else
 	ircd_sendto_chan_local (tm->chan, ":%s!%s@%s KICK %s %s :%s",
-				sender, cl->user, cl->host, chn, lcl, reason);
-#ifdef USE_SERVICES
-      //TODO: inform services
-#endif
+				sender, cl->user, cl->vhost, chn, lcl, reason);
       if ((cmask = strchr (tm->chan->name, ':'))) {
 	cmask++; /* not put '++' into macro below */
 	ircd_sendto_servers_mask_all_ack ((IRCD *)srv->data, tgt, tm->chan,
@@ -865,7 +905,7 @@ static int ircd_kick_sb(INTERFACE *srv, struct peer_t *peer, unsigned short toke
 
 BINDING_TYPE_ircd_server_cmd(ircd_kill_sb);
 static int ircd_kill_sb(INTERFACE *srv, struct peer_t *peer, unsigned short token,
-			const char *sender, const char *lcsender, char *cmd,
+			const char *sender, const char *lcsender,
 			int argc, const char **argv)
 { /* args: <nickname> <comment> */
   CLIENT *cl, *tcl;
@@ -911,6 +951,12 @@ static int ircd_kill_sb(INTERFACE *srv, struct peer_t *peer, unsigned short toke
   if (!CLIENT_IS_REMOTE(tcl))
     New_Request(tcl->via->p.iface, 0, ":%s KILL %s :%s", cl->nick, tcl->nick,
 		reason);		/* notify the victim */
+#ifdef USE_SERVICES
+  ircd_sendto_services_prefix((IRCD *)srv->data, SERVICE_WANT_KILL,
+			      ":%s!%s@%s KILL %s :%s", cl->nick, cl->user,
+			      cl->vhost, tcl->nick, reason);
+  ircd_sendto_services_mark_nick((IRCD *)srv->data, SERVICE_WANT_KILL);
+#endif
   ircd_sendto_servers_all_ack((IRCD *)srv->data, tcl, NULL, pp,
 			      ":%s KILL %s :%s", cl->nick, tcl->nick, reason);
 			      /* broadcast KILL */
@@ -918,7 +964,7 @@ static int ircd_kill_sb(INTERFACE *srv, struct peer_t *peer, unsigned short toke
   tcl->hold_upto = Time + CHASETIMELIMIT; /* make 'nick delay' */
   for (c = NextWord(reason); c > reason && c[-1] != '!'; c--); /* find nick */
   Add_Request(I_PENDING, "*", 0, ":%s!%s@%s QUIT :Killed by %s", tcl->nick,
-	      tcl->user, tcl->host, c);
+	      tcl->user, tcl->vhost, c);
   tcl->host[0] = 0;			/* for collision check */
   Add_Request(I_LOG, "*", F_MODES, "KILL %s :%s", tcl->nick, reason);
   return (1);
@@ -926,7 +972,7 @@ static int ircd_kill_sb(INTERFACE *srv, struct peer_t *peer, unsigned short toke
 
 BINDING_TYPE_ircd_server_cmd(ircd_error_sb);
 static int ircd_error_sb(INTERFACE *srv, struct peer_t *peer, unsigned short token,
-			 const char *sender, const char *lcsender, char *cmd,
+			 const char *sender, const char *lcsender,
 			 int argc, const char **argv)
 { /* args: <error message> */
   /* just broadcasting it to channel and log */
@@ -936,7 +982,7 @@ static int ircd_error_sb(INTERFACE *srv, struct peer_t *peer, unsigned short tok
 
 BINDING_TYPE_ircd_server_cmd(ircd_wallops_sb);
 static int ircd_wallops_sb(INTERFACE *srv, struct peer_t *peer, unsigned short token,
-			   const char *sender, const char *lcsender, char *cmd,
+			   const char *sender, const char *lcsender,
 			   int argc, const char **argv)
 { /* args: <wallops message> */
   struct peer_priv *pp = peer->iface->data; /* it's really peer */
@@ -952,9 +998,6 @@ static int ircd_wallops_sb(INTERFACE *srv, struct peer_t *peer, unsigned short t
     return (1); //TODO: log as duplicate
   /* just broadcast it to everyone */
   ircd_sendto_wallops((IRCD *)srv->data, pp, sender, "%s", argv[0]);
-//  ircd_mark_wallops();
-//  ircd_sendto_servers_all((IRCD *)srv->data, pp, ":%s WALLOPS :%s", sender,
-//			  argv[0]);
   return (1);
 }
 
@@ -972,6 +1015,14 @@ static int _ircd_do_stopic(IRCD *ircd, const char *via, const char *sender,
   //TODO: check permissions may be?
   sz = unistrcut(topic, sizeof(ch->topic), TOPICLEN); /* validate */
   strfcpy(ch->topic, topic, sz+1);
+#ifdef TOPICWHOTIME
+  snprintf(ch->topic_by, sizeof(ch->topic_by), "%s!%s@%s", sender, cl->user,
+	   cl->vhost);
+  ch->topic_since = Time;
+#endif
+#ifdef USE_SERVICES
+  ircd_sendto_services_mark_prefix(ircd, SERVICE_WANT_TOPIC);
+#endif
   if (CLIENT_IS_SERVER(cl)) //TODO: can servers set topics?
     ircd_sendto_chan_local(ch, ":%s TOPIC %s :%s", sender, ch->name, ch->topic);
   else if (CLIENT_IS_SERVICE(cl))
@@ -982,29 +1033,35 @@ static int _ircd_do_stopic(IRCD *ircd, const char *via, const char *sender,
 			   ch->name, ch->topic);
   else
     ircd_sendto_chan_local(ch, ":%s!%s@%s TOPIC %s :%s", sender, cl->user,
-			   cl->host, ch->name, ch->topic);
+			   cl->vhost, ch->name, ch->topic);
   cmask = strchr(ch->name, ':');
   if (cmask)
   {
     cmask++; /* not put '++' into macro below */
+#ifdef USE_SERVICES
+    ircd_sendto_services_mark_nick(ircd, SERVICE_WANT_TOPIC);
+#endif
 #if IRCD_MULTICONNECT
     if (id >= 0) {
-      ircd_sendto_servers_mask_new(ircd, pp, cmask, ":%s ITOPIC %d %s :%s",
-				   sender, id, ch->name, ch->topic);
       ircd_sendto_servers_mask_old(ircd, pp, cmask, ":%s TOPIC %s :%s", sender,
 				   ch->name, ch->topic);
+      ircd_sendto_servers_mask_new(ircd, pp, cmask, ":%s ITOPIC %d %s :%s",
+				   sender, id, ch->name, ch->topic);
     } else
 #endif
       ircd_sendto_servers_mask(ircd, pp, cmask, ":%s TOPIC %s :%s", sender,
 			       ch->name, ch->topic);
     return 1;
   }
+#ifdef USE_SERVICES
+  ircd_sendto_services_mark_nick(ircd, SERVICE_WANT_TOPIC);
+#endif
 #if IRCD_MULTICONNECT
   if (id >= 0) {
-    ircd_sendto_servers_new(ircd, pp, ":%s ITOPIC %d %s :%s",
-			    sender, id, ch->name, ch->topic);
     ircd_sendto_servers_old(ircd, pp, ":%s TOPIC %s :%s",
 			    sender, ch->name, ch->topic);
+    ircd_sendto_servers_new(ircd, pp, ":%s ITOPIC %d %s :%s",
+			    sender, id, ch->name, ch->topic);
   } else
 #endif
     ircd_sendto_servers_all(ircd, pp, ":%s TOPIC %s :%s",
@@ -1015,7 +1072,7 @@ static int _ircd_do_stopic(IRCD *ircd, const char *via, const char *sender,
 #if IRCD_MULTICONNECT
 BINDING_TYPE_ircd_server_cmd(ircd_itopic);
 static int ircd_itopic(INTERFACE *srv, struct peer_t *peer, unsigned short token,
-		       const char *sender, const char *lcsender, char *cmd,
+		       const char *sender, const char *lcsender,
 		       int argc, const char **argv)
 { /* args: <num> <channel> <topic> */
   CLIENT *cl;
@@ -1056,11 +1113,12 @@ static int ircd_itopic(INTERFACE *srv, struct peer_t *peer, unsigned short token
 
 BINDING_TYPE_ircd_server_cmd(ircd_ack);
 static int ircd_ack(INTERFACE *srv, struct peer_t *peer, unsigned short token,
-		    const char *sender, const char *lcsender, char *cmd,
+		    const char *sender, const char *lcsender,
 		    int argc, const char **argv)
 { /* args: <command> <target> [<channel>] */
   register struct peer_priv *pp = peer->iface->data; /* it's really peer */
   const char *channame;
+  ACK *ack = NULL;
 
   if (!(pp->link->cl->umode & A_MULTI)) /* it's ambiguous from RFC2813 server */
     return (0);
@@ -1079,19 +1137,22 @@ static int ircd_ack(INTERFACE *srv, struct peer_t *peer, unsigned short token,
     channame = "0";
   else
     channame = pp->acks->where->name;
-  if (argc == 3 && *argv[2] != '\0' &&
-      (pp->acks->where == NULL || strcmp(argv[2], channame))) {
+  if (argc >= 3 && *argv[2] != '\0' &&
+      (ack = ircd_find_ack(pp, argv[1], argv[2])) == NULL) {
     ERROR("ircd:got ACK %s on %s for unexpected channel %s (expected %s at %s)",
 	  argv[0], argv[1], argv[2], pp->acks->who->nick, channame);
     if (ircd_recover_done(pp, "ACK for unexpected channel") == 0)
       return (0);
-  } else if (argc == 2 &&
-	     (pp->acks->where != NULL || strcmp(argv[1], pp->acks->who->nick))) {
+  } else if (argc == 2 && (ack = ircd_find_ack(pp, argv[1], NULL)) == NULL) {
     ERROR("ircd:got unexpected ACK %s on %s (expected %s %s)", argv[0], argv[1],
-	  pp->acks->who->nick, channame);
+	  pp->acks->who ? pp->acks->who->nick : "(nil)", channame);
     if (ircd_recover_done(pp, "Unexpected ACK arguments") == 0)
       return (0);
   }
+  if (ack == NULL)
+    ack = pp->acks;
+  while (pp->acks != ack)
+    ircd_drop_ack((IRCD *)srv->data, pp);
   ircd_drop_ack((IRCD *)srv->data, pp);
   return (1);
 }
@@ -1122,7 +1183,6 @@ void ircd_server_proto_end (void)
   Delete_Binding ("ircd-server-cmd", (Function)&ircd_ack, NULL);
   _forget_(ACK);
 #endif
-  //TODO: kill messaging iface
 }
 
 void ircd_server_proto_start (void)
@@ -1143,6 +1203,5 @@ void ircd_server_proto_start (void)
 //  Add_Binding ("ircd-server-cmd", "iwallops", 0, 0, (Function)&ircd_iwallops, NULL);
   Add_Binding ("ircd-server-cmd", "ack", 0, 0, (Function)&ircd_ack, NULL);
 #endif
-  //TODO: add messaging iface
 }
 #endif
